# coding: utf-8
# 2021/8/1 @ tongshiwei

import json
from EduNLP.constant import MODEL_DIR
from ..Vector import T2V, get_pretrained_t2v as get_t2v_pretrained_model
from ..Tokenizer import Tokenizer, get_tokenizer
from EduNLP import logger

__all__ = ["I2V", "D2V", "W2V", "get_pretrained_i2v"]


class I2V(object):
    """

    Parameters
    ----------
    tokenizer: str
        the tokenizer name
    t2v: str
        the name of token2vector model
    args:
        the parameters passed to t2v
    tokenizer_kwargs: dict
        the parameters passed to tokenizer
    pretrained_t2v: bool
    kwargs:
        the parameters passed to t2v
    """

    def __init__(self, tokenizer, t2v, *args, tokenizer_kwargs: dict = None, pretrained_t2v=False, **kwargs):

        self.tokenizer: Tokenizer = get_tokenizer(tokenizer, **tokenizer_kwargs if tokenizer_kwargs is not None else {})
        if pretrained_t2v:
            logger.info("Use pretrained t2v model %s" % t2v)
            self.t2v = get_t2v_pretrained_model(t2v, kwargs.get("model_dir", MODEL_DIR))
        else:
            self.t2v = T2V(t2v, *args, **kwargs)
        self.params = {
            "tokenizer": tokenizer,
            "tokenizer_kwargs": tokenizer_kwargs,
            "t2v": t2v,
            "args": args,
            "kwargs": kwargs,
            "pretrained_t2v": pretrained_t2v
        }

    def __call__(self, items, *args, **kwargs):
        return self.infer_vector(items, *args, **kwargs)

    def tokenize(self, items, indexing=True, padding=False, key=lambda x: x, *args, **kwargs) -> list:
        return self.tokenizer(items, key=key, *args, **kwargs)

    def infer_vector(self, items, tokenize=True, indexing=False, padding=False, key=lambda x: x, *args,
                     **kwargs) -> tuple:
        raise NotImplementedError

    def infer_item_vector(self, tokens, *args, **kwargs) -> ...:
        return self.infer_vector(tokens, *args, **kwargs)[0]

    def infer_token_vector(self, tokens, *args, **kwargs) -> ...:
        return self.infer_vector(tokens, *args, **kwargs)[1]

    def save(self, config_path, *args, **kwargs):
        with open(config_path, "w", encoding="utf-8") as wf:
            json.dump(self.params, wf, ensure_ascii=False, indent=2)

    @classmethod
    def load(cls, config_path, *args, **kwargs):
        with open(config_path, encoding="utf-8") as f:
            params: dict = json.load(f)
            tokenizer = params.pop("tokenizer")
            t2v = params.pop("t2v")
            args = params.pop("args")
            kwargs = params.pop("kwargs")
            params.update(kwargs)
            return cls(tokenizer, t2v, *args, **params)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        raise NotImplementedError

    @property
    def vector_size(self):
        return self.t2v.vector_size


class D2V(I2V):
    def infer_vector(self, items, tokenize=True, indexing=False, padding=False, key=lambda x: x, *args,
                     **kwargs) -> tuple:
        tokens = self.tokenize(items, return_token=True, key=key) if tokenize is True else items
        tokens = [token for token in tokens]
        return self.t2v(tokens, *args, **kwargs), None

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        return cls("pure_text", name, pretrained_t2v=True, model_dir=model_dir)


class W2V(I2V):
    def infer_vector(self, items, tokenize=True, indexing=False, padding=False, key=lambda x: x, *args,
                     **kwargs) -> tuple:
        tokens = self.tokenize(items, return_token=True) if tokenize is True else items
        tokens = [token for token in tokens]
        return self.t2v(tokens, *args, **kwargs), self.t2v.infer_tokens(tokens, *args, **kwargs)

    @classmethod
    def from_pretrained(cls, name, model_dir=MODEL_DIR, *args, **kwargs):
        return cls("pure_text", name, pretrained_t2v=True, model_dir=model_dir)


MODELS = {
    "d2v_all_256": [D2V, "d2v_all_256"],
    "d2v_sci_256": [D2V, "d2v_sci_256"],
    "d2v_eng_256": [D2V, "d2v_eng_256"],
    "d2v_lit_256": [D2V, "d2v_lit_256"],
    "w2v_sci_300": [W2V, "w2v_sci_300"],
    "w2v_lit_300": [W2V, "w2v_lit_300"],
}


def get_pretrained_i2v(name, model_dir=MODEL_DIR):
    """

    Parameters
    ----------
    name
    model_dir

    Returns
    -------
    i2v model: I2V

    """
    if name not in MODELS:
        raise KeyError(
            "Unknown model name %s, use one of the provided models: %s" % (name, ", ".join(MODELS.keys()))
        )
    _class, *params = MODELS[name]
    return _class.from_pretrained(*params, model_dir=model_dir)



# coding: utf-8
# 2021/5/18 @ tongshiwei
import logging
import jieba
from .stopwords import DEFAULT_STOPWORDS

jieba.setLogLevel(logging.INFO)


def tokenize(text, granularity="word", stopwords="default"):
    """

    Parameters
    ----------
    text
    granularity
    stopwords: str, None or set

    Returns
    -------

    Examples
    --------
    >>> tokenize("三角函数是基本初等函数之一")
    ['三角函数', '初等', '函数']
    >>> tokenize("三角函数是基本初等函数之一", granularity="char")
    ['三', '角', '函', '数', '基', '初', '函', '数']
    """
    stopwords = DEFAULT_STOPWORDS if stopwords == "default" else stopwords
    stopwords = stopwords if stopwords is not None else {}
    if granularity == "word":
        return [token for token in jieba.cut(text) if token not in stopwords and token.strip()]
    elif granularity == "char":
        stopwords = stopwords if stopwords is not None else {}
        return [token for token in text if token not in stopwords and token.strip()]
    else:
        raise TypeError("Unknown granularity %s" % granularity)



import pytest
from EduNLP.SIF.parser.parser import Parser


def test_parser():
    text = ''
    text_parser = Parser(text)
    text_parser.description_list()

    text = '随机$text{观测}$生产某种零件的A工厂25名工人的日加工零件数_   _'
    text_parser = Parser(text)
    text_parser.description_list()

    text = 'X的分布列为(   )'
    text_parser = Parser(text)
    text_parser.description_list()

    text = '由题意得（ ）'
    text_parser = Parser(text)
    text_parser.description_list()
    assert text_parser.error_flag == 0

    text = '1.命题中真命题的序号是\n ① AB是⊙O的直径，AC是⊙O的切线，BC交⊙O于点E．AC的中点为D'
    text_parser = Parser(text)
    text_parser.description_list()
    assert text_parser.error_flag == 1

    text = r"公式两侧的匹配符号需要完整，如不允许$\frac{y}{x}"
    text_parser = Parser(text)
    text_parser.description_list()
    assert text_parser.error_flag == 1

    text = r"支持公式如$\frac{y}{x}$，$\SIFBlank$，$\FigureID{1}$，不支持公式如$\frac{ \dddot y}{x}$"
    text_parser = Parser(text)
    text_parser.description_list()
    assert text_parser.fomula_illegal_flag == 1



# coding: utf-8
# 2021/5/20 @ tongshiwei

import base64
from io import BytesIO


def image2base64(img):
    buffered = BytesIO()
    img.save(buffered, format="png")
    img_str = base64.b64encode(buffered.getvalue())
    return img_str.decode("utf-8")



# coding: utf-8
# 2021/5/20 @ tongshiwei
import os
import pytest
from PIL import Image
from EduNLP.utils import abs_current_dir, path_append, image2base64


@pytest.fixture(scope="module")
def img_dir():
    return os.path.abspath(path_append(abs_current_dir(__file__), "..", "..", "asset", "_static"))


@pytest.fixture(scope="module")
def figure0(img_dir):
    return Image.open(path_append(img_dir, "item_formula.png", to_str=True))


@pytest.fixture(scope="module")
def figure1(img_dir):
    return Image.open(path_append(img_dir, "item_figure.png", to_str=True))


@pytest.fixture(scope="module")
def figure0_base64(figure0):
    return image2base64(figure0)


@pytest.fixture(scope="module")
def figure1_base64(figure1):
    return image2base64(figure1)



# coding: utf-8
# 2021/5/30 @ tongshiwei

import codecs
import json
import pytest
from EduNLP.utils import abs_current_dir, path_append


@pytest.fixture(scope="module")
def data():
    _data = []
    with codecs.open(path_append(abs_current_dir(__file__), "test.json", to_str=True), encoding="utf-8") as f:
        for line in f.readlines():
            _data.append(json.loads(line))
    return _data



# coding: utf-8
# 2021/5/29 @ tongshiwei
from EduNLP import logger
import multiprocessing
import gensim
from gensim.models import word2vec
from gensim.models.doc2vec import TaggedDocument
from gensim.models.callbacks import CallbackAny2Vec
from EduNLP.SIF.sif import sif4sci
from EduNLP.Vector import D2V, BowLoader
from copy import deepcopy
import itertools as it

__all__ = ["GensimWordTokenizer", "train_vector", "GensimSegTokenizer"]


class GensimWordTokenizer(object):
    """

    Parameters
    ----------
    symbol:
        gm
        fgm
        gmas
        fgmas
    general:
        True when item isn't in standard format, and want to tokenize formulas(except formulas in figure) linearly.
        False when use 'ast' mothed to tokenize formulas instead of 'linear'.

    Returns
    ----------

    Examples
    ----------
    >>> tokenizer = GensimWordTokenizer(symbol="gmas", general=True)
    >>> token_item = tokenizer("有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$")
    >>> print(token_item.tokens[:10])
    ['公式', '[FORMULA]', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[FORMULA]']
    >>> tokenizer = GensimWordTokenizer(symbol="fgmas", general=False)
    >>> token_item = tokenizer("有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$")
    >>> print(token_item.tokens[:10])
    ['公式', '[FORMULA]', '如图', '[FIGURE]', '[FORMULA]', '约束条件', '公式', '[FORMULA]', '[SEP]', '[FORMULA]']
    """
    def __init__(self, symbol="gm", general=False):
        self.symbol = symbol
        if general is True:
            self.tokenization_params = {
                "formula_params": {
                    "method": "linear",
                    "symbolize_figure_formula": True
                }
            }
        else:
            self.tokenization_params = {
                "formula_params": {
                    "method": "ast",
                    "return_type": "list",
                    "ord2token": True
                }
            }

    def batch_process(self, *items):
        pass

    def __call__(self, item):
        return sif4sci(
            item, symbol=self.symbol, tokenization_params=self.tokenization_params, errors="ignore"
        )


class GensimSegTokenizer(object):  # pragma: no cover
    """

    Parameters
    ----------
    symbol:
        gms
        fgm
    """
    def __init__(self, symbol="gms", depth=None, flatten=False, **kwargs):
        self.symbol = symbol
        self.tokenization_params = {
            "formula_params": {
                "method": "ast",
                "return_type": "list",
                "ord2token": True
            }
        }
        self.kwargs = dict(
            add_seg_type=True if depth in {0, 1, 2} else False,
            add_seg_mode="head",
            depth=depth,
            drop="s" if depth not in {0, 1, 2} else ""
        )
        self.kwargs.update(kwargs)
        self.flatten = flatten

    def __call__(self, item, flatten=None, **kwargs):
        flatten = self.flatten if flatten is None else flatten
        tl = sif4sci(
            item, symbol=self.symbol, tokenization_params=self.tokenization_params, errors="ignore"
        )
        if kwargs:
            _kwargs = deepcopy(self.kwargs)
            _kwargs.update(kwargs)
        else:
            _kwargs = self.kwargs
        if tl:
            ret = tl.get_segments(**_kwargs)
            if flatten is True:
                return it.chain(*ret)
            return ret
        return tl


class MonitorCallback(CallbackAny2Vec):
    def __init__(self, test_words):
        self.epoch = 0
        self._test_words = test_words

    def on_epoch_end(self, model):
        logger.info("Epoch #{}: loss-{:.4f} ".format(self.epoch, model.get_latest_training_loss()))
        self.epoch += 1


def train_vector(items, w2v_prefix, embedding_dim=None, method="sg", binary=None, train_params=None):
    monitor = MonitorCallback(["word", "I", "less"])
    _train_params = dict(
        min_count=0,
        vector_size=embedding_dim,
        workers=multiprocessing.cpu_count(),
        callbacks=[monitor]
    )
    if method in {"sg", "cbow"}:
        sg = 1 if method == "sg" else 0
        _train_params["sg"] = sg
        if train_params is not None:
            _train_params.update(train_params)
        model = gensim.models.Word2Vec(
            items, **_train_params
        )
        binary = binary if binary is not None else False
    elif method == "fasttext":
        if train_params is not None:
            _train_params.update(train_params)
        model = gensim.models.FastText(
            sentences=items,
            **_train_params
        )
        binary = binary if binary is not None else True
    elif method == "d2v":
        if train_params is not None:
            _train_params.update(train_params)
        docs = [TaggedDocument(doc, [i]) for i, doc in enumerate(items)]
        model = gensim.models.Doc2Vec(
            docs, **_train_params
        )
        binary = binary if binary is not None else True
    elif method == "bow":
        model = gensim.corpora.Dictionary(items)
        binary = binary if binary is not None else True
    elif method == "tfidf":
        dictionary_path = train_vector(items, w2v_prefix, method="bow")
        dictionary = BowLoader(dictionary_path)
        corpus = [dictionary.infer_vector(item) for item in items]
        model = gensim.models.TfidfModel(corpus)
        binary = binary if binary is not None else True
    else:
        raise ValueError("Unknown method: %s" % method)

    filepath = w2v_prefix + method
    if embedding_dim is not None:
        filepath = filepath + "_" + str(embedding_dim)

    if binary is True:
        filepath += ".bin"
        logger.info("model is saved to %s" % filepath)
        model.save(filepath)
    else:
        if method in {"fasttext", "d2v"}:  # pragma: no cover
            logger.warning("binary should be True for %s, otherwise all vectors for ngrams will be lost." % method)
        filepath += ".kv"
        logger.info("model is saved to %s" % filepath)
        model.wv.save(filepath)
    return filepath



from EduNLP.Formula.ast import str2ast, katex_parse


class Parser:
    def __init__(self, data):
        self.lookahead = 0
        self.head = 0
        self.text = data
        self.error_message = ''
        self.error_postion = 0
        self.error_flag = 0
        self.modify_flag = 0
        self.warnning = 0
        self.fomula_illegal_flag = 0
        self.fomula_illegal_message = ''

        # 定义特殊变量
        self.len_bracket = len('$\\SIFChoice$')
        self.len_underline = len('$\\SIFBlank$')

        # 定义 token
        self.error = -1
        self.character = 1
        self.en_pun = 2
        self.ch_pun = 3
        self.latex = 4
        self.end = 5
        self.empty = 6
        self.modify = 7
        self.blank = 8

        self.en_pun_list = [',', '.', '?', '!',
                            ':', ';', '\'', '\"', '(', ')', ' ', '_', '/', '|', '\\', '<', '>', '[', ']',
                            '-']  # add some other chars
        self.ch_pun_list = ['，', '。', '！', '？', '：',
                            '；', '‘', '’', '“', '”', '（', '）', ' ', '、', '《', '》', '—', '．']
        self.in_list = [',', '_', '-', '%']
        self.flag_list = ['，', '。', '！', '？', '：',
                          '；', '‘', '’', '“', '”', '（', '）', ' ', '、', '《', '》',
                          '$', ',', '.', '?', '!', ':', ';', '\'', '\"', '(', ')', ' ', '_', '/', '|', '<', '>', '-',
                          '[', ']', '—']

    def is_number(self, uchar):
        """判断一个unicode是否是数字"""
        if u'\u0030' <= uchar <= u'\u0039':
            # print(uchar, ord(uchar))(u'\u0030' <= uchar <= u'\u0039')
            return True
        else:
            return False

    def is_alphabet(self, uchar):
        """判断一个unicode是否是英文字母"""
        if (u'\u0041' <= uchar <= u'\u005a') or (u'\u0061' <= uchar <= u'\u007a'):
            return True
        else:
            return False

    def is_chinese(self, uchar):
        """判断一个unicode是否是汉字"""
        if u'\u4e00' <= uchar <= u'\u9fa5':
            return True
        else:
            return False

    def _is_formula_legal(self, formula_str):
        r"""
        Judge whether the current formula meet our specification or not.

        Parameters
        ----------
        formula_str

        Returns
        -------
        True or False

        """
        legal_tags = ['FormFigureID', 'FormFigureBase64', 'FigureID', 'FigureBase64',
                      'SIFBlank', 'SIFChoice', 'SIFTag', 'SIFSep', 'SIFUnderline', 'textf']
        for tag in legal_tags:
            if tag in formula_str:
                return True
        try:
            katex_parse(formula_str)
        except Exception as e:
            assert 'ParseError' in str(e)
            self.fomula_illegal_message = "[FormulaError] " + str(e)
            self.fomula_illegal_flag = 1
            return False
        return True

    def call_error(self):
        """语法解析函数"""
        # print('ERROR::position is >>> ',self.head)
        # print('ERROR::match is >>>', self.text[self.head])
        self.error_postion = self.head
        self.error_message = self.text[:self.head + 1]
        self.error_flag = 1

    def get_token(self):
        if self.head >= len(self.text):
            return self.empty
        ch = self.text[self.head]
        if self.is_chinese(ch):
            # 匹配中文字符 [\u4e00-\u9fa5]
            self.head += 1
            return self.character
        elif self.is_alphabet(ch):
            # 匹配公式之外的英文字母，只对两个汉字之间的字母做修正，其余匹配到的情况视为不合 latex 语法录入的公式
            left = head = self.head
            if self.head == 0:
                while (head < len(self.text) and (
                        self.is_alphabet(self.text[head]) or self.text[head] in self.in_list)):
                    head += 1
                if head == len(self.text) or self.is_chinese(self.text[head]) or self.text[head] in self.flag_list:
                    self.head = head
                    self.text = self.text[:left] + "$" + self.text[left:head] + "$" + self.text[head:]
                    self.head += 2
                    #                     print(self.text[left:self.head])
                    self.modify = 1
                    return self.modify
            else:
                forward = self.text[self.head - 1]
                if self.is_chinese(forward) or forward in self.flag_list:
                    while (head < len(self.text) and (
                            self.is_alphabet(self.text[head]) or self.text[head] in self.in_list)):
                        head += 1
                    if head == len(self.text) or self.is_chinese(self.text[head]) or self.text[head] in self.flag_list:
                        self.head = head
                        self.text = self.text[:left] + "$" + self.text[left:head] + "$" + self.text[head:]
                        self.head += 2
                        self.modify_flag = 1
                        return self.modify
            # self.call_error()
            # return self.error

        elif self.is_number(ch):
            # 匹配公式之外的数字，只对两个汉字之间的数字做修正，其余匹配到的情况视为不合 latex 语法录入的公式
            left = head = self.head
            if self.head == 0:
                while (head < len(self.text) and (
                        self.is_number(self.text[head]) or self.text[head] in self.in_list)):
                    head += 1
                if head == len(self.text) or self.is_chinese(self.text[head]) or self.text[head] in self.flag_list:
                    self.head = head
                    self.text = self.text[:left] + "$" + self.text[left:head] + "$" + self.text[head:]
                    self.head += 2
                    self.modify_flag = 1
                    return self.modify

            else:
                forward = self.text[self.head - 1]
                if self.is_chinese(forward) or forward in self.flag_list:
                    while (head < len(self.text) and (
                            self.is_number(self.text[head]) or self.text[head] in self.in_list)):
                        head += 1

                    if head == len(self.text) or self.is_chinese(self.text[head]) or self.text[head] in self.flag_list:
                        self.head = head
                        self.text = self.text[:left] + "$" + self.text[left:head] + "$" + self.text[head:]
                        self.head += 2
                        self.modify_flag = 1
                        return self.modify
            # self.call_error()
            # return self.error

        elif ch == '\n':
            # 匹配换行符
            self.head += 1
            return self.end

        elif ch in self.ch_pun_list:
            # 匹配中文标点
            left = self.head
            self.head += 1
            if self.text[left] == '（':
                # 匹配到一个左括号
                while self.text[self.head] == ' ' or self.text[self.head] == '\xa0':
                    self.head += 1
                if self.text[self.head] == '）':
                    self.head += 1
                    self.text = self.text[:left] + '$\\SIFChoice$' + self.text[self.head:]
                    self.head += self.len_bracket
                    self.modify_flag = 1
                    return self.modify
            return self.ch_pun
        elif ch in self.en_pun_list:
            # 匹配英文标点
            # print('en-pun-list')
            left = self.head
            self.head += 1
            if self.text[left] == '(':
                # 匹配到一个左括号
                while self.text[self.head] == ' ' or self.text[self.head] == '\xa0':
                    self.head += 1
                if self.text[self.head] == ')':
                    self.head += 1
                    self.text = self.text[:left] + '$\\SIFChoice$' + self.text[self.head:]
                    self.head += self.len_bracket
                    self.modify_flag = 1
                    return self.modify
            if self.text[left] == '_':
                # 匹配到一个下划线
                # print('this is an underline')
                while self.text[self.head] == '_' or self.text[self.head] == ' ':
                    self.head += 1
                    if self.head >= len(self.text):
                        break
                # print('change the text')
                self.text = self.text[:left] + '$\\SIFBlank$' + self.text[self.head:]
                self.head += self.len_underline
                # print(self.text)
                self.modify_flag = 1
                return self.modify
            return self.en_pun

        elif ch == '$':
            # 匹配 latex 公式
            self.head += 1
            flag = 1
            formula_start = self.head
            while self.head < len(self.text) and self.text[self.head] != '$':
                ch_informula = self.text[self.head]
                if flag and self.is_chinese(ch_informula):
                    # latex 中出现中文字符，打印且只打印一次 warning
                    print("Warning: there is some chinese characters in formula!")
                    self.warnning = 1
                    flag = 0
                self.head += 1
            if self.head >= len(self.text):
                self.call_error()
                return self.error
            # 检查latex公式的完整性和可解析性
            if not self._is_formula_legal(self.text[formula_start:self.head]):
                self.call_error()
                return self.error
            self.head += 1
            # print('is latex!')
            return self.latex
        else:
            self.call_error()
            return self.error

    def next_token(self):
        #         print('call next_token')
        #         if self.error_flag:
        #             return
        self.lookahead = self.get_token()
        if self.error_flag:
            return

    def match(self, terminal):
        #         print('call match')
        # if self.error_flag:
        #     return
        if self.lookahead == terminal:
            self.next_token()
            if self.error_flag:
                return
        # else:
        #     print('match error!')
        #     self.call_error()

    def txt(self):
        #         print('call txt')
        #         if self.error_flag:
        #             return
        self.lookahead = self.get_token()
        if self.error_flag:
            return
        if self.lookahead == self.character or self.lookahead == self.en_pun or \
                self.lookahead == self.ch_pun or self.lookahead == self.latex:
            self.match(self.lookahead)

    def txt_list(self):
        #         print('call txt_list')
        #         if self.error_flag:
        #             return
        self.txt()
        if self.error_flag:
            return
        if self.lookahead != self.empty:
            self.txt_list()

    def description(self):
        #         print('call description')
        #         if self.error_flag:
        #             return
        self.txt_list()
        if self.error_flag:
            return
        if self.lookahead == self.empty:
            self.match(self.lookahead)

    def description_list(self):
        r"""
        use Parser to process and describe the txt

        Parameters
        ----------

        Returns
        ----------

        Examples
        --------
        >>> text = '生产某种零件的A工厂25名工人的日加工零件数_   _'
        >>> text_parser = Parser(text)
        >>> text_parser.description_list()
        >>> text_parser.text
        '生产某种零件的$A$工厂$25$名工人的日加工零件数$\\SIFBlank$'
        >>> text = 'X的分布列为(   )'
        >>> text_parser = Parser(text)
        >>> text_parser.description_list()
        >>> text_parser.text
        '$X$的分布列为$\\SIFChoice$'
        >>> text = '① AB是⊙O的直径，AC是⊙O的切线，BC交⊙O于点E．AC的中点为D'
        >>> text_parser = Parser(text)
        >>> text_parser.description_list()
        >>> text_parser.error_flag
        1
        >>> text = '支持公式如$\\frac{y}{x}$，$\\SIFBlank$，$\\FigureID{1}$，不支持公式如$\\frac{ \\dddot y}{x}$'
        >>> text_parser = Parser(text)
        >>> text_parser.description_list()
        >>> text_parser.fomula_illegal_flag
        1
        """
        # print('call description_list')
        self.description()
        if self.error_flag:
            # print("Error")
            return
        if self.lookahead != self.empty:
            self.description_list()  # pragma: no cover
        else:
            self.error_flag = 0
            # print('parse successfully!')



# coding: utf-8
# 2021/3/8 @ tongshiwei

import matplotlib.pyplot as plt
from sklearn.tree._export import _MPLTreeExporter
from sklearn.tree._reingold_tilford import buchheim, Tree
from matplotlib.text import Annotation


class TreePlotter(_MPLTreeExporter):
    def recurse(self, node, ax, scale_x, scale_y, height, depth=0):
        kwargs = dict(bbox=self.bbox_args, ha='center', va='center',
                      zorder=100 - 10 * depth, xycoords='axes pixels')

        if self.fontsize is not None:
            kwargs['fontsize'] = self.fontsize

        # offset things by .5 to center them in plot
        xy = ((node.x + .5) * scale_x, height - (node.y + .5) * scale_y)

        if self.max_depth is None or depth <= self.max_depth:
            # if self.filled:
            #     kwargs['bbox']['fc'] = self.get_fill_color(tree,
            #                                                node.tree.node_id)
            if node.parent is None:
                # root
                ax.annotate(node.tree.label, xy, **kwargs)
            else:
                xy_parent = ((node.parent.x + .5) * scale_x,
                             height - (node.parent.y + .5) * scale_y)
                kwargs["arrowprops"] = self.arrow_args
                ax.annotate(node.tree.label, xy_parent, xy, **kwargs)
            for child in node.children:
                self.recurse(child, ax, scale_x, scale_y, height,
                             depth=depth + 1)

        else:
            xy_parent = ((node.parent.x + .5) * scale_x,
                         height - (node.parent.y + .5) * scale_y)
            kwargs["arrowprops"] = self.arrow_args
            kwargs['bbox']['fc'] = 'grey'
            ax.annotate("\n  (...)  \n", xy_parent, xy, **kwargs)

    def _make_forest(self, ast):
        forest = []
        for node in ast:
            if node["structure"]["father"] is None:
                return Tree()
            else:
                pass

        return Tree(name, node_id, *children)

    def export(self, formula_ast, ax=None):
        self.filled = False

        if ax is None:
            ax = plt.gca()
        ax.clear()
        ax.set_axis_off()
        # my_tree = self._make_tree(0, decision_tree.tree_,
        #                           decision_tree.criterion)
        my_tree = self._make_forest(formula_ast)
        draw_tree = buchheim(my_tree)

        # important to make sure we're still
        # inside the axis after drawing the box
        # this makes sense because the width of a box
        # is about the same as the distance between boxes
        max_x, max_y = draw_tree.max_extents() + 1
        ax_width = ax.get_window_extent().width
        ax_height = ax.get_window_extent().height

        scale_x = ax_width / max_x
        scale_y = ax_height / max_y

        self.recurse(draw_tree, ax,
                     scale_x, scale_y, ax_height)

        anns = [ann for ann in ax.get_children()
                if isinstance(ann, Annotation)]

        # update sizes of all bboxes
        renderer = ax.figure.canvas.get_renderer()

        for ann in anns:
            ann.update_bbox_position_size(renderer)

        if self.fontsize is None:
            # get figure to data transform
            # adjust fontsize to avoid overlap
            # get max box width and height
            extents = [ann.get_bbox_patch().get_window_extent()
                       for ann in anns]
            max_width = max([extent.width for extent in extents])
            max_height = max([extent.height for extent in extents])
            # width should be around scale_x in axis coordinates
            size = anns[0].get_fontsize() * min(scale_x / max_width,
                                                scale_y / max_height)
            for ann in anns:
                ann.set_fontsize(size)

        return anns



# coding: utf-8
# 2021/8/2 @ tongshiwei

import fire

from EduNLP.I2V.i2v import MODELS


def list_i2v():
    print("\n".join(MODELS.keys()))


def cli():  # pragma: no cover
    fire.Fire({"i2v": list_i2v})



# coding: utf-8
# 2021/5/18 @ tongshiwei

import os
from EduNLP.utils import abs_current_dir, path_append

DEFAULT_FILEPATH = os.path.abspath(
    path_append(abs_current_dir(__file__), "..", "..", "..", "meta_data", "sif_stopwords.txt")
)


def get_stopwords(filepath=DEFAULT_FILEPATH):
    _stopwords = set()
    with open(filepath, encoding="utf-8") as f:
        for line in f:
            _stopwords.add(line.strip())

    return _stopwords


DEFAULT_STOPWORDS = get_stopwords()



# coding: utf-8
# 2021/7/13 @ tongshiwei

class Vector(object):
    def infer_vector(self, items, *args, **kwargs) -> ...:
        pass

    def infer_tokens(self, items, *args, **kwargs) -> ...:
        pass

    @property
    def vector_size(self):
        raise NotImplementedError

    @property
    def is_frozen(self):  # pragma: no cover
        return True

    def freeze(self, *args, **kwargs):  # pragma: no cover
        pass



# coding: utf-8
# 2021/5/29 @ tongshiwei
import logging


def get_logger():
    _logger = logging.getLogger("EduNLP")
    _logger.setLevel(logging.INFO)
    _logger.propagate = False
    ch = logging.StreamHandler()
    ch.setFormatter(logging.Formatter('%(name)s, %(levelname)s %(message)s'))
    ch.setLevel(logging.INFO)
    _logger.addHandler(ch)
    return _logger


logger = get_logger()



# coding: utf-8
# 2021/3/8 @ tongshiwei

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from sklearn.tree._export import _MPLTreeExporter
from sklearn.tree._reingold_tilford import buchheim, Tree
from matplotlib.text import Annotation


class TreePlotter(_MPLTreeExporter):
    def recurse(self, node, ax, scale_x, scale_y, height, scale_x_offset=0, scale_y_offset=0, depth=0):
        kwargs = dict(bbox=self.bbox_args, ha='center', va='center',
                      zorder=100 - 10 * depth, xycoords='axes pixels')

        if self.fontsize is not None:
            kwargs['fontsize'] = self.fontsize

        # offset things by .5 to center them in plot
        xy = ((node.x + .5 + scale_x_offset) * scale_x, height - (node.y + .5 + scale_y_offset) * scale_y)

        if self.max_depth is None or depth <= self.max_depth:
            # if self.filled:
            #     kwargs['bbox']['fc'] = self.get_fill_color(tree,
            #                                                node.tree.node_id)
            if node.parent is None:
                # root
                ax.annotate(node.tree.label, xy, **kwargs)
            else:
                xy_parent = ((node.parent.x + .5 + scale_x_offset) * scale_x,
                             height - (node.parent.y + .5 + scale_y_offset) * scale_y)
                kwargs["arrowprops"] = self.arrow_args
                ax.annotate(node.tree.label, xy_parent, xy, **kwargs)
            for child in node.children:
                self.recurse(child, ax, scale_x, scale_y, height,
                             scale_x_offset, scale_y_offset, depth=depth + 1)

        else:
            xy_parent = ((node.parent.x + .5) * scale_x,
                         height - (node.parent.y + .5) * scale_y)
            kwargs["arrowprops"] = self.arrow_args
            kwargs['bbox']['fc'] = 'grey'
            ax.annotate("\n  (...)  \n", xy_parent, xy, **kwargs)

    def node_to_str(self, node: dict):
        return "\n".join(["%s: %s" % (k, v) for k, v in node.items()])

    def make_tree(self, ast: nx.DiGraph, root=None):
        if root is None:
            assert nx.is_tree(ast)
            for node in ast.nodes:
                if not list(ast.predecessors(node)):
                    root = node
                    break

        assert root is not None

        return self._make_tree(
            root, ast
        )

    def _make_tree(self, node, ast):
        return Tree(
            self.node_to_str(ast.nodes[node]),
            node,
            *[self._make_tree(successor, ast) for successor in ast.successors(node)]
        )

    def export(self, formula_ast, ax=None):
        self.filled = False

        if ax is None:
            ax = plt.gca()
        ax.clear()
        ax.set_axis_off()
        # my_tree = self._make_tree(0, decision_tree.tree_,
        #                           decision_tree.criterion)
        my_tree = self.make_tree(formula_ast)
        draw_tree = buchheim(my_tree)

        # important to make sure we're still
        # inside the axis after drawing the box
        # this makes sense because the width of a box
        # is about the same as the distance between boxes
        max_x, max_y = draw_tree.max_extents() + 1
        ax_width = ax.get_window_extent().width
        ax_height = ax.get_window_extent().height

        scale_x = ax_width / max_x
        scale_y = ax_height / max_y

        self.recurse(draw_tree, ax,
                     scale_x, scale_y, ax_height)

        anns = [ann for ann in ax.get_children()
                if isinstance(ann, Annotation)]

        # update sizes of all bboxes
        renderer = ax.figure.canvas.get_renderer()

        for ann in anns:
            ann.update_bbox_position_size(renderer)

        if self.fontsize is None:
            # get figure to data transform
            # adjust fontsize to avoid overlap
            # get max box width and height
            extents = [ann.get_bbox_patch().get_window_extent()
                       for ann in anns]
            max_width = max([extent.width for extent in extents])
            max_height = max([extent.height for extent in extents])
            # width should be around scale_x in axis coordinates
            size = anns[0].get_fontsize() * min(scale_x / max_width,
                                                scale_y / max_height)
            for ann in anns:
                ann.set_fontsize(size)

        return anns


class ForestPlotter(TreePlotter):
    def export(self, forest, ax=None, root_list=None):
        self.filled = False

        if ax is None:
            ax = plt.gca()
        ax.clear()
        ax.set_axis_off()
        # my_tree = self._make_tree(0, decision_tree.tree_,
        #                           decision_tree.criterion)
        my_forest = []
        if root_list is None:
            if isinstance(forest, list):
                for tree in forest:
                    my_tree = self.make_tree(tree)
                    draw_tree = buchheim(my_tree)
                    my_forest.append(draw_tree)
        else:
            for tree in root_list:
                my_tree = self.make_tree(forest, tree)
                draw_tree = buchheim(my_tree)
                my_forest.append(draw_tree)

        # important to make sure we're still
        # inside the axis after drawing the box
        # this makes sense because the width of a box
        # is about the same as the distance between boxes
        offset = np.stack([draw_tree.max_extents() + 1 for draw_tree in my_forest])
        max_x, max_y = sum(offset[:, 0]) + 1, max(offset[:, 1]) + 1
        ax_width = ax.get_window_extent().width
        ax_height = ax.get_window_extent().height

        scale_x = ax_width / max_x
        scale_y = ax_height / max_y

        x_offset = np.cumsum(offset[:, 0])
        x_offset = [0] + x_offset.tolist()
        for i, draw_tree in enumerate(my_forest):
            self.recurse(
                draw_tree, ax, scale_x, scale_y,
                ax_height,
                x_offset[i]
            )

        anns = [ann for ann in ax.get_children() if isinstance(ann, Annotation)]

        # update sizes of all bboxes
        renderer = ax.figure.canvas.get_renderer()

        for ann in anns:
            ann.update_bbox_position_size(renderer)

        if self.fontsize is None:
            # get figure to data transform
            # adjust fontsize to avoid overlap
            # get max box width and height
            extents = [ann.get_bbox_patch().get_window_extent()
                       for ann in anns]
            max_width = max([extent.width for extent in extents])
            max_height = max([extent.height for extent in extents])
            # width should be around scale_x in axis coordinates
            size = anns[0].get_fontsize() * min(scale_x / max_width,
                                                scale_y / max_height)
            for ann in anns:
                ann.set_fontsize(size)

        return anns


#
# import networkx as nx
#
# g = nx.DiGraph()
# g.add_node(0, value=1, id=0)
# g.add_node(1, value=2, id=1)
# g.add_node(2, id=2)
# g.add_node(3, id=3)
# g.add_node(4, id=4)
# g.add_edge(0, 1)
# g.add_edge(0, 2)
# g.add_edge(1, 3)
# g.add_edge(3, 4)
#
# g2 = nx.DiGraph()
# g2.add_node(5, value=10, id=5)
# g2.add_node(6, value=200, id=6)
# g2.add_node(7, id=7)
# g2.add_edge(5, 6)
# g2.add_edge(5, 7)
# # ForestPlotter().export([g, g2])
# TreePlotter().export(g)
# plt.show()



# coding: utf-8
# 2021/5/18 @ tongshiwei
import base64
import numpy as np
import re
from contextlib import contextmanager
from ..constants import Symbol, TEXT_SYMBOL, FORMULA_SYMBOL, FIGURE_SYMBOL, QUES_MARK_SYMBOL, TAG_SYMBOL, SEP_SYMBOL


class TextSegment(str):
    pass


class LatexFormulaSegment(str):
    pass


class Figure(object):
    def __init__(self, is_base64=False):
        self.base64 = is_base64
        self.figure = None

    @classmethod
    def base64_to_numpy(cls, figure: str):
        return np.frombuffer(base64.b64decode(figure), dtype=np.uint8)


class FigureFormulaSegment(Figure):
    def __init__(self, src, is_base64=False, figure_instance: (dict, bool) = None):
        super(FigureFormulaSegment, self).__init__(is_base64)
        self.src = src
        if self.base64 is True:
            self.figure = self.src[len(r"\FormFigureBase64") + 1: -1]
            if figure_instance is True or (isinstance(figure_instance, dict) and figure_instance.get("base64") is True):
                self.figure = self.base64_to_numpy(self.figure)
        else:
            self.figure = self.src[len(r"\FormFigureID") + 1: -1]
            if isinstance(figure_instance, dict):
                self.figure = figure_instance[self.figure]

    def __repr__(self):
        if self.base64 is True:
            return FORMULA_SYMBOL
        return str(self.src)


class FigureSegment(Figure):
    def __init__(self, src, is_base64=False, figure_instance: (dict, bool) = None):
        super(FigureSegment, self).__init__(is_base64)
        self.src = src
        if self.base64 is True:
            self.figure = self.src[len(r"\FigureBase64") + 1: -1]
            if figure_instance is True or (isinstance(figure_instance, dict) and figure_instance.get("base64") is True):
                self.figure = self.base64_to_numpy(self.figure)
        else:
            self.figure = self.src[len(r"\FigureID") + 1: -1]
            if isinstance(figure_instance, dict):
                self.figure = figure_instance[self.figure]

    def __repr__(self):
        if self.base64 is True:
            return FIGURE_SYMBOL
        return str(self.src)


class QuesMarkSegment(str):
    pass


class TagSegment(str):
    pass


class SepSegment(str):
    pass


class SegmentList(object):
    def __init__(self, item, figures: dict = None):
        self._segments = []
        self._text_segments = []
        self._formula_segments = []
        self._figure_segments = []
        self._ques_mark_segments = []
        self._tag_segments = []
        self._sep_segments = []

        # remove $\textf{*} from the item$
        item_no_textf = "".join(re.split(r"\$\\textf\{([^,]+?),b?d?i?t?u?w?}\$", item))

        segments = re.split(r"(\$.+?\$)", item_no_textf)
        for segment in segments:
            if not segment:
                continue
            if not re.match(r"\$.+?\$", segment):
                self.append(TextSegment(segment))
            elif re.match(r"\$\\FormFigureID\{.+?}\$", segment):
                self.append(FigureFormulaSegment(segment[1:-1], is_base64=False, figure_instance=figures))
            elif re.match(r"\$\\FormFigureBase64\{.+?}\$", segment):
                self.append(FigureFormulaSegment(segment[1:-1], is_base64=True, figure_instance=figures))
            elif re.match(r"\$\\FigureID\{.+?}\$", segment):
                self.append(FigureSegment(segment[1:-1], is_base64=False, figure_instance=figures))
            elif re.match(r"\$\\FigureBase64\{.+?}\$", segment):
                self.append(FigureSegment(segment[1:-1], is_base64=True, figure_instance=figures))
            elif re.match(r"\$\\(SIFBlank|SIFChoice)\$", segment):
                self.append(QuesMarkSegment(segment[1:-1]))
            elif re.match(r"\$\\SIFTag\{.+?}\$", segment):
                self.append(TagSegment(segment[1:-1]))
            elif re.match(r"\$\\SIFSep\$", segment):
                self.append(SepSegment(segment[1:-1]))
            else:
                self.append(LatexFormulaSegment(segment[1:-1]))
        self._seg_idx = None

    def __repr__(self):
        return str(self.segments)

    def __len__(self):
        return len(self._segments)

    def append(self, segment) -> None:
        if isinstance(segment, TextSegment):
            self._text_segments.append(len(self))
        elif isinstance(segment, (LatexFormulaSegment, FigureFormulaSegment)):
            self._formula_segments.append(len(self))
        elif isinstance(segment, FigureSegment):
            self._figure_segments.append(len(self))
        elif isinstance(segment, QuesMarkSegment):
            self._ques_mark_segments.append(len(self))
        elif isinstance(segment, TagSegment):
            self._tag_segments.append(len(self))
        elif isinstance(segment, SepSegment):
            self._sep_segments.append(len(self))
        else:
            raise TypeError("Unknown Segment Type: %s" % type(segment))
        self._segments.append(segment)

    @property
    def segments(self):
        if self._seg_idx is None:
            return self._segments
        else:
            return [s for i, s in enumerate(self._segments) if i in self._seg_idx]

    @property
    def text_segments(self):
        return [self._segments[i] for i in self._text_segments]

    @property
    def formula_segments(self):
        return [self._segments[i] for i in self._formula_segments]

    @property
    def figure_segments(self):
        return [self._segments[i] for i in self._figure_segments]

    @property
    def ques_mark_segments(self):
        return [self._segments[i] for i in self._ques_mark_segments]

    @property
    def tag_segments(self):
        return [self._segments[i] for i in self._tag_segments]

    def to_symbol(self, idx, symbol):
        self._segments[idx] = symbol

    def symbolize(self, to_symbolize="fgm"):
        """

        Parameters
        ----------
        to_symbolize:
            "t": text
            "f": formula
            "g": figure
            "m": question mark

        Returns
        -------

        """
        if "t" in to_symbolize:
            for idx in self._text_segments:
                self.to_symbol(idx, Symbol(TEXT_SYMBOL))
        if "f" in to_symbolize:
            for idx in self._formula_segments:
                self.to_symbol(idx, Symbol(FORMULA_SYMBOL))
        if "g" in to_symbolize:
            for idx in self._figure_segments:
                self.to_symbol(idx, Symbol(FIGURE_SYMBOL))
        if "m" in to_symbolize:
            for idx in self._ques_mark_segments:
                self.to_symbol(idx, Symbol(QUES_MARK_SYMBOL))
        if "a" in to_symbolize:
            for idx in self._tag_segments:
                self.to_symbol(idx, Symbol(TAG_SYMBOL))
        if "s" in to_symbolize:
            for idx in self._sep_segments:
                self.to_symbol(idx, Symbol(SEP_SYMBOL))

    @contextmanager
    def filter(self, drop: (set, str) = "", keep: (set, str) = "*"):
        _drop = {c for c in drop} if isinstance(drop, str) else drop
        if keep == "*":
            _keep = {c for c in "tfgmas" if c not in _drop}
        else:
            _keep = {c for c in keep if c not in _drop} if isinstance(keep, str) else keep
        self._seg_idx = set()
        if "t" in _keep:
            self._seg_idx |= set(self._text_segments)
        if "f" in _keep:
            self._seg_idx |= set(self._formula_segments)
        if "g" in _keep:
            self._seg_idx |= set(self._figure_segments)
        if "m" in _keep:
            self._seg_idx |= set(self._ques_mark_segments)
        if "a" in _keep:
            self._seg_idx |= set(self._tag_segments)
        if "s" in _keep:
            self._seg_idx |= set(self._sep_segments)
        yield
        self._seg_idx = None

    def describe(self):
        return {
            "t": len(self._text_segments),
            "f": len(self._formula_segments),
            "g": len(self._figure_segments),
            "m": len(self._ques_mark_segments),
        }


def seg(item, figures=None, symbol=None):
    r"""

    Parameters
    ----------
    item
    figures
    symbol

    Returns
    -------

    Examples
    --------
    >>> test_item = r"如图所示，则$\bigtriangleup ABC$的面积是$\SIFBlank$。$\FigureID{1}$"
    >>> s = seg(test_item)
    >>> s
    ['如图所示，则', '\\bigtriangleup ABC', '的面积是', '\\SIFBlank', '。', \FigureID{1}]
    >>> s.describe()
    {'t': 3, 'f': 1, 'g': 1, 'm': 1}
    >>> with s.filter("f"):
    ...     s
    ['如图所示，则', '的面积是', '\\SIFBlank', '。', \FigureID{1}]
    >>> with s.filter(keep="t"):
    ...     s
    ['如图所示，则', '的面积是', '。']
    >>> with s.filter():
    ...     s
    ['如图所示，则', '\\bigtriangleup ABC', '的面积是', '\\SIFBlank', '。', \FigureID{1}]
    >>> seg(test_item, symbol="fgm")
    ['如图所示，则', '[FORMULA]', '的面积是', '[MARK]', '。', '[FIGURE]']
    >>> seg(test_item, symbol="tfgm")
    ['[TEXT]', '[FORMULA]', '[TEXT]', '[MARK]', '[TEXT]', '[FIGURE]']
    >>> seg(r"如图所示，则$\FormFigureID{0}$的面积是$\SIFBlank$。$\FigureID{1}$")
    ['如图所示，则', \FormFigureID{0}, '的面积是', '\\SIFBlank', '。', \FigureID{1}]
    >>> seg(r"如图所示，则$\FormFigureID{0}$的面积是$\SIFBlank$。$\FigureID{1}$", symbol="fgm")
    ['如图所示，则', '[FORMULA]', '的面积是', '[MARK]', '。', '[FIGURE]']
    >>> s.text_segments
    ['如图所示，则', '的面积是', '。']
    >>> s.formula_segments
    ['\\bigtriangleup ABC']
    >>> s.figure_segments
    [\FigureID{1}]
    >>> s.ques_mark_segments
    ['\\SIFBlank']
    >>> test_item_1 = {
    ...     "stem": r"若复数$z=1+2 i+i^{3}$，则$|z|=$",
    ...     "options": ['0', '1', r'$\sqrt{2}$', '2']
    ... }
    >>> from EduNLP.utils import dict2str4sif
    >>> test_item_1_str = dict2str4sif(test_item_1)
    >>> test_item_1_str  # doctest: +ELLIPSIS
    '$\\SIFTag{stem_begin}$...$\\SIFTag{stem_end}$$\\SIFTag{options_begin}$$\\SIFTag{list_0}$0...$\\SIFTag{options_end}$'
    >>> s1 = seg(test_item_1_str, symbol="tfgm")
    >>> s1  # doctest: +ELLIPSIS
    ['\\SIFTag{stem_begin}'...'\\SIFTag{stem_end}', '\\SIFTag{options_begin}', '\\SIFTag{list_0}', ...]
    >>> with s1.filter(keep="a"):
    ...     s1  # doctest: +ELLIPSIS
    [...'\\SIFTag{list_0}', '\\SIFTag{list_1}', '\\SIFTag{list_2}', '\\SIFTag{list_3}', '\\SIFTag{options_end}']
    >>> s1.tag_segments  # doctest: +ELLIPSIS
    ['\\SIFTag{stem_begin}', '\\SIFTag{stem_end}', '\\SIFTag{options_begin}', ... '\\SIFTag{options_end}']
    >>> test_item_1_str_2 = dict2str4sif(test_item_1, tag_mode="head", add_list_no_tag=False)
    >>> seg(test_item_1_str_2, symbol="tfgmas")  # doctest: +ELLIPSIS
    ['[TAG]', ... '[TAG]', '[TEXT]', '[SEP]', '[TEXT]', '[SEP]', '[FORMULA]', '[SEP]', '[TEXT]']
    >>> s2 = seg(test_item_1_str_2, symbol="fgm")
    >>> s2.tag_segments
    ['\\SIFTag{stem}', '\\SIFTag{options}']
    >>> test_item_2 = r"已知$y=x$，则以下说法中$\textf{正确,b}$的是"
    >>> s2 = seg(test_item_2)
    >>> s2.text_segments
    ['已知', '，则以下说法中正确的是']
    """
    segments = SegmentList(item, figures)
    if symbol is not None:
        segments.symbolize(symbol)
    return segments



# coding: utf-8
# 2021/5/20 @ tongshiwei
import networkx as nx
from EduNLP.Formula import Formula


# def inorder_traversal(ast: nx.DiGraph):
#     visit = set()
#     nodes = []
#
#     def _inorder_traversal(_node):
#         if _node in visit:
#             return
#         successors = list(ast.successors(_node))
#         if successors:
#             if len(successors) <= 2:
#                 _inorder_traversal(successors[0])
#                 nodes.append(_node)
#                 visit.add(_node)
#                 if len(successors) == 2:
#                     _inorder_traversal(successors[1])
#             else:
#                 nodes.append(_node)
#                 for successor in successors:
#                     if successor in visit:
#                         continue
#                     _inorder_traversal(successor)
#         else:
#             nodes.append(_node)
#
#     for node in ast.nodes:
#         if node in visit or list(ast.predecessors(node)):
#             continue
#         _inorder_traversal(node)
#     return nodes

def traversal_formula(ast, ord2token=False, var_numbering=False, strategy="post", *args, **kwargs):
    tokens = []
    if strategy == "post":
        order = nx.dfs_postorder_nodes(ast)
    elif strategy == "linear":  # pragma: no cover
        order = ast.nodes
    else:  # pragma: no cover
        raise ValueError("Unknown traversal strategy: %s" % strategy)
    for i in order:
        node = ast.nodes[i]
        if node["type"] == "ignore":
            continue
        if ord2token is True and node["type"] in ["mathord", "textord", "text"]:
            if var_numbering is True and node["type"] == "mathord":
                tokens.append("%s_%s" % (node["type"], node.get("var", "con")))
            else:
                tokens.append(node["type"])
        else:
            tokens.append(node["text"])
    return tokens


def ast_tokenize(formula, ord2token=False, var_numbering=False, return_type="formula", *args, **kwargs):
    """

    Parameters
    ----------
    formula
    ord2token
    var_numbering
    return_type
    args
    kwargs

    Returns
    -------

    Examples
    --------
    >>> ast_tokenize(r"{x + y}^\\frac{\\pi}{2} + 1 = x", return_type="list")
    ['x', '+', 'y', '{ }', '\\\\pi', '{ }', '2', '{ }', '\\\\frac', '\\\\supsub', '+', '1', '=', 'x']
    >>> ast_tokenize(r"{x + y}^\\frac{\\pi}{2} + 1 = x", return_type="list", ord2token=True)
    ['mathord', '+', 'mathord', '{ }', 'mathord', '{ }', 'textord', '{ }', '\\\\frac', '\\\\supsub', '+', 'textord', \
'=', 'mathord']
    >>> ast_tokenize(r"{x + y}^\\frac{\\pi}{2} + 1 = x", return_type="list", ord2token=True, var_numbering=True)
    ['mathord_0', '+', 'mathord_1', '{ }', 'mathord_con', '{ }', 'textord', '{ }', '\\\\frac', '\\\\supsub', \
'+', 'textord', '=', 'mathord_0']
    >>> len(ast_tokenize(r"{x + y}^\\frac{\\pi}{2} + 1 = x", return_type="ast").nodes)
    14
    >>> ast_tokenize(r"{x + y}^\\frac{\\pi}{2} + 1 = x")
    <Formula: {x + y}^\\frac{\\pi}{2} + 1 = x>
    """
    if return_type == "list":
        ast = Formula(formula, variable_standardization=True).ast_graph
        return traversal_formula(ast, ord2token=ord2token, var_numbering=var_numbering)
    elif return_type == "formula":
        return Formula(formula)
    elif return_type == "ast":
        return Formula(formula).ast_graph
    else:
        raise ValueError()


if __name__ == '__main__':
    print(ast_tokenize(r"{x + y}^\frac{\pi}{2} + 1 = x", return_type="list", ord2token=True, var_numbering=True))



# coding: utf-8
# 2021/3/8 @ tongshiwei
# This code is modified from https://scikit-learn.org/stable/modules/generated/sklearn.tree.plot_tree.html

import matplotlib.pyplot as plt
from matplotlib.text import Annotation
from .utils import buchheim


class Tree(object):
    def __init__(self, label="", node_id=-1, *children):
        self.label = label
        self.node_id = node_id
        if children:
            self.children = children
        else:
            self.children = []


class TreePlotter(object):
    def __init__(self, max_depth=None, feature_names=None,
                 class_names=None, label='all', filled=False,
                 impurity=True, node_ids=False,
                 proportion=False, rotate=False, rounded=False,
                 precision=3, fontsize=None):

        self.fontsize = fontsize

        # The depth of each node for plotting with 'leaf' option
        self.ranks = {'leaves': []}
        # The colors to render each node with
        self.colors = {'bounds': None}

        self.characters = ['#', '[', ']', '<=', '\n', '', '']

        self.bbox_args = dict(fc='w')
        if self.rounded:
            self.bbox_args['boxstyle'] = "round"

        self.arrow_args = dict(arrowstyle="<-")

    def _make_tree(self, node_id, et, criterion, depth=0):



        name = self.node_to_str(et, node_id, criterion=criterion)
        if (et.children_left[node_id] != _tree.TREE_LEAF
                and (self.max_depth is None or depth <= self.max_depth)):
            children = [self._make_tree(et.children_left[node_id], et,
                                        criterion, depth=depth + 1),
                        self._make_tree(et.children_right[node_id], et,
                                        criterion, depth=depth + 1)]
        else:
            return Tree(name, node_id)
        return Tree(name, node_id, *children)

    def export(self, decision_tree, ax=None):
        import matplotlib.pyplot as plt
        from matplotlib.text import Annotation

        if ax is None:
            ax = plt.gca()
        ax.clear()
        ax.set_axis_off()
        my_tree = self._make_tree(0, decision_tree.tree_,
                                  decision_tree.criterion)
        draw_tree = buchheim(my_tree)

        # important to make sure we're still
        # inside the axis after drawing the box
        # this makes sense because the width of a box
        # is about the same as the distance between boxes
        max_x, max_y = draw_tree.max_extents() + 1
        ax_width = ax.get_window_extent().width
        ax_height = ax.get_window_extent().height

        scale_x = ax_width / max_x
        scale_y = ax_height / max_y

        self.recurse(draw_tree, decision_tree.tree_, ax,
                     scale_x, scale_y, ax_height)

        anns = [ann for ann in ax.get_children()
                if isinstance(ann, Annotation)]

        # update sizes of all bboxes
        renderer = ax.figure.canvas.get_renderer()

        for ann in anns:
            ann.update_bbox_position_size(renderer)

        if self.fontsize is None:
            # get figure to data transform
            # adjust fontsize to avoid overlap
            # get max box width and height
            extents = [ann.get_bbox_patch().get_window_extent()
                       for ann in anns]
            max_width = max([extent.width for extent in extents])
            max_height = max([extent.height for extent in extents])
            # width should be around scale_x in axis coordinates
            size = anns[0].get_fontsize() * min(scale_x / max_width,
                                                scale_y / max_height)
            for ann in anns:
                ann.set_fontsize(size)

        return anns

    def recurse(self, node, tree, ax, scale_x, scale_y, height, depth=0):
        kwargs = dict(bbox=self.bbox_args, ha='center', va='center',
                      zorder=100 - 10 * depth, xycoords='axes pixels')

        if self.fontsize is not None:
            kwargs['fontsize'] = self.fontsize

        # offset things by .5 to center them in plot
        xy = ((node.x + .5) * scale_x, height - (node.y + .5) * scale_y)

        if self.max_depth is None or depth <= self.max_depth:
            if self.filled:
                kwargs['bbox']['fc'] = self.get_fill_color(tree,
                                                           node.tree.node_id)
            if node.parent is None:
                # root
                ax.annotate(node.tree.label, xy, **kwargs)
            else:
                xy_parent = ((node.parent.x + .5) * scale_x,
                             height - (node.parent.y + .5) * scale_y)
                kwargs["arrowprops"] = self.arrow_args
                ax.annotate(node.tree.label, xy_parent, xy, **kwargs)
            for child in node.children:
                self.recurse(child, tree, ax, scale_x, scale_y, height,
                             depth=depth + 1)

        else:
            xy_parent = ((node.parent.x + .5) * scale_x,
                         height - (node.parent.y + .5) * scale_y)
            kwargs["arrowprops"] = self.arrow_args
            kwargs['bbox']['fc'] = 'grey'
            ax.annotate("\n  (...)  \n", xy_parent, xy, **kwargs)




# coding: utf-8
# 2021/5/20 @ tongshiwei
from typing import List, Dict
from .katex import katex


__all__ = ["str2ast", "get_edges", "ast", "link_variable", "katex_parse"]


def katex_parse(formula):
    return katex.katex.__parse(formula,{'displayMode':True,'trust': True}).to_list()


def str2ast(formula: str, *args, **kwargs):
    return ast(formula, is_str=True, *args, **kwargs)


def ast(formula: (str, List[Dict]), index=0, forest_begin=0, father_tree=None, is_str=False):
    """
    The origin code author is https://github.com/hxwujinze

    Parameters
    ----------
    formula: str or List[Dict]
        公式字符串或通过katex解析得到的结构体
    index: int
        本子树在树上的位置
    forest_begin: int
        本树在森林中的起始位置
    father_tree: List[Dict]
        父亲树
    is_str: bool


    Returns
    ----------
    tree: List[Dict]
        重新解析形成的特征树

    todo: finish all types

    Notes
    ----------
    Some functions are not supportd in ``katex``
    e.g.,

    1. tag
        - ``\\begin{equation} \\tag{tagName} F=ma \\end{equation}``
        - ``\\begin{align} \\tag{1} y=x+z \\end{align}``
        - ``\\tag*{hi} x+y^{2x}``
    2. dddot
        - ``\\frac{ \\dddot y }{ x }``

    For more information, refer to
    `katex support table <https://github.com/KaTeX/KaTeX/blob/master/docs/support_table.md>`_
    """
    tree = []
    index += forest_begin
    json_ast: List[Dict] = katex.katex.__parse(formula,{'displayMode':True,'trust': True}).to_list() if is_str else formula
    last_node = None

    for item in json_ast:
        private_index = len(tree)
        role = None
        if 'role' in item:
            role = item['role']
        tree_node = {
            'val': {'id': private_index + index, 'type': None, 'text': None, 'role': role},
            'structure': {'bro': [None, None], 'child': None, 'father': None, 'forest': None}
        }

        tree_node['val']['type'] = item['type']

        if index > forest_begin:
            tree_node['structure']['father'] = index - 1
        if tree_node['val']['type'] == "mathord" or tree_node['val']['type'] == "textord":
            tree_node['val']['text'] = item['text'].replace('\\prime', '’')
            tree.append(tree_node)
        elif tree_node['val']['type'] == "atom":
            tree_node['val']['text'] = item['text'].replace('\\cdotp', '·')
            tree_node['val']['type'] = item['family']
            tree.append(tree_node)

        elif tree_node['val']['type'] == "op":
            tree_node['val']['text'] = "\\op" if 'name' not in item else item['name']
            if item['symbol'] and 'body' in item:
                tree_node['structure']['child'] = [1 + private_index + index]
                tree.append(tree_node)
                tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
            else:
                tree_node['val']['text'] = item['name']
                tree.append(tree_node)
        elif tree_node['val']['type'] == "genfrac":
            item['numer']['role'] = 'numer'
            item['denom']['role'] = 'denom'
            tree_node['val']['text'] = '\\frac'
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast([item['numer'], item['denom']], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'sqrt':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = '\\sqrt'
            tree.append(tree_node)
            item['body']['role'] = 'body'
            if item['index']:
                item['index']['role'] = 'index'
                tree += ast([item['body'], item['index']], index=len(tree) + index, father_tree=tree)
            else:
                tree += ast([item['body']], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'array':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['arraydims'] = item['arraystretch']
            tree_node['val']['text'] = '\\begin {matrix} \\end {matrix}'
            tree.append(tree_node)
            bodys = []
            for litem in item['body']:
                for citem in litem:
                    citem['role'] = 'body'
                    bodys.append(citem)
            tree += ast(bodys, index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'styling':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            for citem in item['body']:
                citem['role'] = 'body'
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'xArrow':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item['label']
            tree.append(tree_node)
            item['body']['role'] = 'body'
            item['below']['role'] = 'below'
            tree += ast([item['body'], item['below']], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'overline':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = '\\' + tree_node['val']['type']
            tree.append(tree_node)
            item['body']['role'] = 'body'
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == "accent":
            tree_node['val']['text'] = item['label']
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            item['base']['role'] = 'base'
            tree += ast([item['base']], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == "supsub":
            _tree = []
            if 'base' in item and item['base'] is not None:
                item['base']['role'] = 'base'
                _tree.append(item['base'])
            if 'sub' in item and item['sub']:
                item['sub']['role'] = 'sub'
                _tree.append(item['sub'])
            if 'sup' in item and item['sup']:
                item['sup']['role'] = 'sup'
                _tree.append(item['sup'])

            tree_node['val']['text'] = "\\supsub"
            if _tree != []:
                tree_node['structure']['child'] = [1 + private_index + index]
                tree.append(tree_node)
                tree += ast(_tree, index=len(tree) + index, father_tree=tree)
            else:
                tree.append(tree_node)

        elif tree_node['val']['type'] == "ordgroup":
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = '{ }'
            tree.append(tree_node)
            for citem in item['body']:
                citem['role'] = 'body'
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == "mclass":
            tree_node['val']['text'] = item['mclass']
            for citem in item['body']:
                citem['role'] = 'body'
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] == 'leftright':
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item['left']
            tree_node['val']['right'] = item['right']
            tree.append(tree_node)
            for citem in item['body']:
                citem['role'] = 'body'
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)

        elif tree_node['val']['type'] in {"kern"}:
            # \quad
            tree_node['val']['text'] = tree_node['val']['type']
            tree_node['val']['type'] = "ignore"
            tree.append(tree_node)

        elif tree_node['val']['type'] == "text":
            # \text{}
            tree_node['val']['text'] = "".join([e['text'] for e in item["body"]])
            tree.append(tree_node)
        # --------------------- new node --------------------- # 
        elif tree_node['val']['type'] == "size":
            # nknown usage : different from "sizing"
            continue
        elif tree_node['val']['type'] == "internal": 
            # unknown usage
            continue
        elif tree_node['val']['type'] == "cr":
            # new line
            continue
        elif tree_node['val']['type'] == "infix":
            continue
        elif tree_node['val']['type'] == "rule":
            # ignore layout setting
            continue
        elif tree_node['val']['type'] == "cdlabel":
            tree_node['val']['text'] = item['side']
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)      
            item['label']['role'] = 'label'
            tree += ast([item['label']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "cdlabelparent":
            tree_node['val']['text'] = "\\cdlabelparent"
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            item['fragment']['role'] = 'fragment'
            tree += ast([item['fragment']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "color":
            tree_node['val']['text'] = "\\color"
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "color-token":
            tree_node['val']['text'] = "\\color-token"
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "raw":
            tree_node['val']['text'] = item['string']
            tree.append(tree_node)
        elif tree_node['val']['type'] == "styling":
            # to be confirmed
            tree_node['val']['text'] = "\\styling" 
            # tree_node['val']['text'] = item["style"] ！= None ? item["style"]: "\\styling"
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "tag":
            continue
            # not supported in Katex yet
            # tree_node['structure']['child'] = [1 + private_index + index]
            # tree_node['val']['text'] = '\\tag' # equations with order number
            # tree.append(tree_node)
            # body_item = {'type':'nodelist','role': 'body','body': item['body']}
            # tag_item = {'type':'nodelist','role': 'tag','body': item['tag']}
            # tree += ast([body_item, tag_item], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "verb":
            tree_node['val']['text'] = item['body'] # "original copy", source code
            tree.append(tree_node)
        elif tree_node['val']['type'] in ["spacing","accent-token","op-token"]:
            tree_node['val']['text'] = item['text']
            tree.append(tree_node) 
        elif tree_node['val']['type'] in ["accent","accentUnder"]: 
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item["label"]
            tree.append(tree_node)
            item['base']['role'] = 'base'
            tree += ast([item['base']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "delimsizing":
            # contains symbols for size settings, including "(",")", etc
            tree_node['val']['text'] = item['delim']
            tree.append(tree_node)
        elif tree_node['val']['type'] == "enclose":
            # setting deleting line effect
            tree_node['val']['text'] = item['label']
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "environment":
            tree_node['val']['text'] = item['name']
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            item['nameGroup']['role'] = 'nameGroup'
            tree += ast([item['nameGroup']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "url":
            # continue
            tree_node['val']['text'] = item['url']
            tree.append(tree_node)
        elif tree_node['val']['type'] == "href":
            # continue
            tree_node['val']['text'] = item['href']
            tree_node['structure']['child'] = [1 + private_index + index]
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "html":
            # continue
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\html"  
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "htmlmathml":
            # continue
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\htmlmathml"
            tree.append(tree_node)
            html_item = {'type':'nodelist','role': 'html','body': item['html']} # ?
            mathml_item = {'type':'nodelist','role': 'mathml','body': item['mathml']} # ?
            tree += ast([html_item,mathml_item], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "includegraphics":
            # continue
            tree_node['val']['text'] = item['src']
            tree.append(tree_node)
        elif tree_node['val']['type'] == "font":
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item["font"]  # font name
            tree.append(tree_node)
            item['body']['role'] = 'body'
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "hbox":
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = '\\hbox' # box layout
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "vcenter":
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = '\\vcenter' # box layout
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "horizBrace":
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item['label']
            tree.append(tree_node)
            item['base']['role'] = 'base'
            tree += ast([item['base']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "lap":
            # layout setting (overlap) 
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = item["alignment"] # methods of overlap (llap | rlap)
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "sizing":
            # consider ignoring size
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\sizing"
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "smash":
            # layout setting : smash (height | width)
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\smash"
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "mathchoice":
            # provides content that is dependent on the current style (display, text, script, or scriptscript).
            # eg: \mathchoice {#1}{#2}{#3}{#4}
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\mathchoice"
            tree.append(tree_node)
            mathchoiceList = []
            for choice in ["display","text","script","scriptscript"]:
                citem = {'type':'nodelist','role':choice, 'body':item[choice]}
                mathchoiceList.append(citem)
            tree += ast(mathchoiceList, index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "operatorname":
            # unknown usage
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\operatorname"
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] in ["overline","underline"]:
            # consider ignoring line
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\" + tree_node['val']['type']
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "raisebox":
            # raise or lower the height of the text
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\raisebox"
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == "leftright-right":
            # paired with leftright
            tree_node['val']['text'] = item["delim"]
            tree.append(tree_node)
        elif tree_node['val']['type'] == "middle":
            # symbols with height setting, such as "|"
            tree_node['val']['text'] = item["delim"]
            tree.append(tree_node)
        elif tree_node['val']['type'] in ["phantom","hphantom","vphantom"]:
            # set space distance by the length of content
            # continue 
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\" + tree_node['val']['type']
            tree.append(tree_node)
            tree += ast([item['body']], index=len(tree) + index, father_tree=tree)
        elif tree_node['val']['type'] == 'nodelist':
            # process node list specially
            tree_node['structure']['child'] = [1 + private_index + index]
            tree_node['val']['text'] = "\\" + item["role"]
            tree.append(tree_node)
            tree += ast(item['body'], index=len(tree) + index, father_tree=tree)
        
        else:
            tree_node['structure']['child'] = [1 + private_index + index]

            if "text" in item:
                tree_node['val']['text'] = item["text"]
            else:
                tree_node['val']['text'] = item["type"]
            tree_node['val']['type'] = "other"
            tree.append(tree_node)
            Role = ['body', 'base', 'sup', 'sub', 'numer', 'denom', 'index', 'below','nameGroup','fragment','label', 'other']
            childrole = []
            for role_item in Role:
                if role_item in item:
                    if role_item == "body" and isinstance(item[role_item], dict) is False:
                        # \text{}
                        childrole.extend(item[role_item])
                    else:
                        item[role_item]['role'] = role_item
                        childrole.append(item[role_item])
            tree += ast(childrole, index=len(tree) + index, father_tree=tree)
        if item:
            if item != json_ast[0]:
                tree[private_index]['structure']['bro'][0] = last_node + index
                last_node = private_index
            else:
                last_node = private_index
            if item != json_ast[-1]:
                tree[private_index]['structure']['bro'][1] = len(tree) + index
                if father_tree:
                    father_tree[tree[private_index]['structure']['father'] - index]['structure']['child'].append(
                        len(tree) + index
                    )

    return tree


def link_variable(forest):
    """
    建森林

    Parameters
    ----------
    forest: List[Dict]

    Returns
    -------
    trees: List[Dict]

    """
    forest_connect_dict = {}
    for node in forest:
        if node['val']['type'] == 'mathord':
            if node['val']['text'] not in forest_connect_dict:
                forest_connect_dict[node['val']['text']] = []
            forest_connect_dict[node['val']['text']].append(forest.index(node))

    for k, v in forest_connect_dict.items():
        if len(v) > 1:
            for i in range(0, len(v)):
                l_v = [] + v
                index = l_v.pop(i)
                forest[index]['structure']['forest'] = l_v
    variable_connect_dict = {
        "var2id": forest_connect_dict,
        "var_code": {}
    }
    return variable_connect_dict


def get_edges(forest):
    """
    构造边集合

    Parameters
    ----------
    forest: List[Dict]
        森林

    Returns
    ----------
    edges: list of tuple(src,dst,type)
        边集合
    """
    edges = []
    for node in forest:
        index = node["val"]["id"]
        edges.append((index, index, 1))
        if node['structure']['bro'][1] is not None:
            edges.append((index, node['structure']['bro'][1], 2))
        if node['structure']['bro'][0] is not None:
            edges.append((index, node['structure']['bro'][0], 2))
        if node['structure']['child'] is not None:
            for item in node['structure']['child']:
                edges.append((index, item, 3))
        if node['structure']['father'] is not None:
            edges.append((index, node['structure']['father'], 4))
        if node['structure']['forest'] is not None:
            for item in node['structure']['forest']:
                edges.append((index, item, 5))
    return edges



# coding: utf-8
# 2021/7/12 @ tongshiwei

import torch
from torch import nn
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence
from baize.torch import load_net


class LM(nn.Module):
    """
    Examples
    --------
    >>> import torch
    >>> seq_idx = torch.LongTensor([[1, 2, 3], [1, 2, 0], [3, 0, 0]])
    >>> seq_len = torch.LongTensor([3, 2, 1])
    >>> lm = LM("RNN", 4, 3, 2)
    >>> output, hn = lm(seq_idx, seq_len)
    >>> output.shape
    torch.Size([3, 3, 2])
    >>> hn.shape
    torch.Size([1, 3, 2])
    >>> lm = LM("RNN", 4, 3, 2, num_layers=2)
    >>> output, hn = lm(seq_idx, seq_len)
    >>> output.shape
    torch.Size([3, 3, 2])
    >>> hn.shape
    torch.Size([2, 3, 2])
    """

    def __init__(self, rnn_type: str, vocab_size: int, embedding_dim: int, hidden_size: int, num_layers=1,
                 bidirectional=False, embedding=None, model_params=None, **kwargs):
        super(LM, self).__init__()
        rnn_type = rnn_type.upper()
        self.embedding = torch.nn.Embedding(vocab_size, embedding_dim) if embedding is None else embedding
        self.c = False
        if rnn_type == "RNN":
            self.rnn = torch.nn.RNN(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
        elif rnn_type == "LSTM":
            self.rnn = torch.nn.LSTM(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
            self.c = True
        elif rnn_type == "GRU":
            self.rnn = torch.nn.GRU(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
        elif rnn_type == "ELMO":
            bidirectional = True
            self.rnn = torch.nn.LSTM(
                embedding_dim, hidden_size, num_layers, bidirectional=bidirectional, **kwargs
            )
            self.c = True
        else:
            raise TypeError("Unknown rnn_type %s" % rnn_type)

        self.num_layers = num_layers
        self.bidirectional = bidirectional
        if bidirectional is True:
            self.num_layers *= 2
        self.hidden_size = hidden_size

        if model_params:
            load_net(model_params, self, allow_missing=True)

    def forward(self, seq_idx, seq_len):
        seq = self.embedding(seq_idx)
        pack = pack_padded_sequence(seq, seq_len, batch_first=True)
        h0 = torch.zeros(self.num_layers, seq.shape[0], self.hidden_size)
        if self.c is True:
            c0 = torch.zeros(self.num_layers, seq.shape[0], self.hidden_size)
            output, (hn, _) = self.rnn(pack, (h0, c0))
        else:
            output, hn = self.rnn(pack, h0)
        output, _ = pad_packed_sequence(output, batch_first=True)
        return output, hn



# coding: utf-8
# 2021/5/20 @ tongshiwei

import pytest
from EduNLP.SIF.constants import Symbol
from EduNLP.SIF.segment.segment import SegmentList
from EduNLP.SIF.tokenization import text
from EduNLP.SIF.tokenization import formula
from EduNLP.SIF.tokenization.tokenization import TokenList


def test_text_tokenization():
    with pytest.raises(TypeError):
        text.tokenize("12345", "alpha")


def test_formula_tokenization():
    with pytest.raises(ValueError):
        formula.ast_token.ast_tokenize("1 + 1", return_type="graph")

    with pytest.raises(TypeError):
        formula.tokenize("1 + 1", method="plain")

    # with pytest.raises(TypeError):
    #     formula.tokenize(r"\phantom{=}56+4", method="ast")


def test_tokenization():
    tl = TokenList(SegmentList(""))
    with pytest.raises(TypeError):
        tl.append(Symbol("[Unknown]"))

    with pytest.raises(TypeError):
        tl.append("[Unknown]")



# coding: utf-8
# 2021/5/18 @ tongshiwei

from contextlib import contextmanager
from EduNLP.Formula import link_formulas as _link_formulas, Formula
from ..constants import (
    Symbol, TEXT_SYMBOL, FIGURE_SYMBOL, FORMULA_SYMBOL, QUES_MARK_SYMBOL, TAG_SYMBOL, SEP_SYMBOL,
    TEXT_BEGIN, TEXT_END, FORMULA_BEGIN, FORMULA_END
)
from ..segment import (SegmentList, TextSegment, FigureSegment, LatexFormulaSegment, FigureFormulaSegment,
                       QuesMarkSegment, Figure, TagSegment, SepSegment)
from . import text, formula

__all__ = ["TokenList", "tokenize", "link_formulas"]


class TokenList(object):
    """
    Attributes
    -------------

    """
    def __init__(self, segment_list: SegmentList, text_params=None, formula_params=None, figure_params=None):
        self._tokens = []
        self._text_tokens = []
        self._formula_tokens = []
        self._figure_tokens = []
        self._ques_mark_tokens = []
        self._tag_tokens = []
        self._sep_tokens = []
        self._segments = []
        self._seg_types = {
            "t": [],
            "f": [],
            "g": [],
            "m": [],
            "a": [],
            "s": []
        }
        self.text_params = text_params if text_params is not None else {}
        self.symbolize_figure_formula = False
        self.skip_figure_formula = False
        if formula_params is not None:
            if "symbolize_figure_formula" in formula_params:
                self.symbolize_figure_formula = formula_params.pop("symbolize_figure_formula")
            if "skip_figure_formula" in formula_params:
                self.skip_figure_formula = formula_params.pop("skip_figure_formula")

        self.formula_params = formula_params if formula_params is not None else {"method": "linear"}
        self.formula_tokenize_method = self.formula_params.get("method")
        self.figure_params = figure_params if figure_params is not None else {}
        self.extend(segment_list.segments)
        self._token_idx = None

    def _variable_standardization(self):
        if self.formula_tokenize_method == "ast":
            ast_formulas = [self._tokens[i] for i in self._formula_tokens if isinstance(self._tokens[i], Formula)]
            if ast_formulas:
                _link_formulas(*ast_formulas, link_vars=self.formula_params.get("var_numbering", False))

    @contextmanager
    def add_seg_type(self, seg_type, tar: list, add_seg_type=True, mode="delimiter"):
        if add_seg_type is True and mode in {"delimiter", "head"}:
            if seg_type == "t":
                tar.append(TEXT_BEGIN)
            elif seg_type == "f" and (
                    self.formula_params.get("method") == "ast" and self.formula_params.get("return_type", "list")
            ):
                tar.append(FORMULA_BEGIN)
        yield
        if add_seg_type is True and mode in {"delimiter", "tail"}:
            if seg_type == "t":
                tar.append(TEXT_END)
            elif seg_type == "f" and (
                    self.formula_params.get("method") == "ast" and self.formula_params.get("return_type", "list")
            ):
                tar.append(FORMULA_END)

    def get_segments(self, add_seg_type=True, add_seg_mode="delimiter", keep="*", drop="",
                     depth=None):  # pragma: no cover
        r"""

        Parameters
        ----------
        add_seg_type
        add_seg_mode:
            delimiter: both in the head and at the tail
            head: only in the head
            tail: only at the tail
        keep
        drop
        depth: int or None
            0: only separate at \SIFSep
            1: only separate at \SIFTag
            2: separate at \SIFTag and \SIFSep
            otherwise, separate all segments

        Returns
        -------

        """
        keep = set("tfgmas" if keep == "*" else keep) - set(drop)
        _segments = []
        _segment = []
        close_tag = False
        for start, end, seg_type in self._segments:
            if depth == 0:
                if seg_type == "s":
                    close_tag = True
            elif depth == 1:
                if seg_type == "a":
                    close_tag = True
            elif depth == 2:
                if seg_type in {"s", "a"}:
                    close_tag = True
            else:
                close_tag = True
            if seg_type in keep:
                with self.add_seg_type(seg_type, _segment, add_seg_type, add_seg_mode):
                    for token in self._tokens[start: end]:
                        self.__add_token(token, _segment)
            if close_tag is True and _segment:
                _segments.append(_segment)
                _segment = []
        return _segments

    def __get_segments(self, seg_type):
        _segments = []
        for i in self._seg_types[seg_type]:
            _segment = []
            start, end, _ = self._segments[i]
            for token in self._tokens[start: end]:
                self.__add_token(token, _segment)
            if _segment:
                _segments.append(_segment)
        return _segments

    @property
    def text_segments(self):
        return self.__get_segments("t")

    @property
    def formula_segments(self):
        return self.__get_segments("f")

    @property
    def figure_segments(self):
        return self.__get_segments("g")

    @property
    def ques_mark_segments(self):
        return self.__get_segments("m")

    @property
    def tokens(self):
        tokens = []
        if self._token_idx is not None:
            for i, token in enumerate(self._tokens):
                if i in self._token_idx:
                    self.__add_token(token, tokens)
        else:
            for token in self._tokens:
                self.__add_token(token, tokens)
        return tokens

    def append_text(self, segment, symbol=False):
        with self._append("t"):
            if symbol is False:
                tokens = text.tokenize(segment, **self.text_params)
                for token in tokens:
                    self._text_tokens.append(len(self._tokens))
                    self._tokens.append(token)
            else:
                self._text_tokens.append(len(self._tokens))
                self._tokens.append(segment)

    def append_formula(self, segment, symbol=False, init=True):
        with self._append("f"):
            if symbol is True:
                self._formula_tokens.append(len(self._tokens))
                self._tokens.append(segment)
            elif self.skip_figure_formula and isinstance(segment, FigureFormulaSegment):
                # skip the FigureFormulaSegment
                pass
            elif self.symbolize_figure_formula and isinstance(segment, FigureFormulaSegment):
                self._formula_tokens.append(len(self._tokens))
                self._tokens.append(Symbol(FORMULA_SYMBOL))
            elif isinstance(segment, FigureFormulaSegment):
                self._formula_tokens.append(len(self._tokens))
                self._tokens.append(segment)
            elif self.formula_params.get("method") == "ast":
                self._formula_tokens.append(len(self._tokens))
                self._tokens.append(Formula(segment, init=init))
            else:
                tokens = formula.tokenize(segment, **self.formula_params)
                for token in tokens:
                    self._formula_tokens.append(len(self._tokens))
                    self._tokens.append(token)

    def append_figure(self, segment, **kwargs):
        with self._append("g"):
            self._figure_tokens.append(len(self._tokens))
            self._tokens.append(segment)

    def append_ques_mark(self, segment, **kwargs):
        with self._append("m"):
            self._ques_mark_tokens.append(len(self._tokens))
            self._tokens.append(segment)

    def append_tag(self, segment, **kwargs):
        with self._append("a"):
            self._tag_tokens.append(len(self._tokens))
            self._tokens.append(segment)

    def append_sep(self, segment, **kwargs):
        with self._append("s"):
            self._sep_tokens.append(len(self._tokens))
            self._tokens.append(segment)

    @contextmanager
    def _append(self, seg_type):
        start = len(self._tokens)
        yield
        end = len(self._tokens)
        self._seg_types[seg_type].append(len(self._segments))
        self._segments.append((start, end, seg_type))

    def append(self, segment, lazy=False):
        if isinstance(segment, TextSegment):
            self.append_text(segment)
        elif isinstance(segment, (LatexFormulaSegment, FigureFormulaSegment)):
            self.append_formula(segment, init=not lazy)
            if lazy is False:
                self._variable_standardization()
        elif isinstance(segment, FigureSegment):
            self.append_figure(segment)
        elif isinstance(segment, QuesMarkSegment):
            self.append_ques_mark(segment)
        elif isinstance(segment, TagSegment):
            self.append_tag(segment)
        elif isinstance(segment, SepSegment):
            self.append_sep(segment)
        elif isinstance(segment, Symbol):
            if segment == TEXT_SYMBOL:
                self.append_text(segment, symbol=True)
            elif segment == FORMULA_SYMBOL:
                self.append_formula(segment, symbol=True)
            elif segment == FIGURE_SYMBOL:
                self.append_figure(segment, symbol=True)
            elif segment == QUES_MARK_SYMBOL:
                self.append_ques_mark(segment, symbol=True)
            elif segment == TAG_SYMBOL:
                self.append_tag(segment, symbol=True)
            elif segment == SEP_SYMBOL:
                self.append_sep(segment, symbol=True)
            else:
                raise TypeError("Unknown symbol type: %s" % segment)
        else:
            raise TypeError("Unknown segment type: %s" % type(segment))

    def extend(self, segments):
        for segment in segments:
            self.append(segment, True)
        self._variable_standardization()

    @property
    def text_tokens(self):
        return [self._tokens[i] for i in self._text_tokens]

    def __add_token(self, token, tokens):
        if isinstance(token, Formula):
            if self.formula_params.get("return_type") == "list":
                tokens.extend(formula.traversal_formula(token.ast_graph, **self.formula_params))
            elif self.formula_params.get("return_type") == "ast":
                tokens.append(token.ast_graph)
            else:
                tokens.append(token)
        elif isinstance(token, Figure):
            if self.figure_params.get("figure_instance") is True:
                tokens.append(token.figure)
            else:
                tokens.append(token)
        else:
            tokens.append(token)

    @property
    def formula_tokens(self):
        tokens = []
        for i in self._formula_tokens:
            self.__add_token(self._tokens[i], tokens)
        return tokens

    @property
    def figure_tokens(self):
        tokens = []
        for i in self._figure_tokens:
            self.__add_token(self._tokens[i], tokens)
        return tokens

    @property
    def ques_mark_tokens(self):
        return [self._tokens[i] for i in self._ques_mark_tokens]

    def __repr__(self):
        return str(self.tokens)

    @property
    def inner_formula_tokens(self):
        return [self._tokens[i] for i in self._formula_tokens]

    @contextmanager
    def filter(self, drop: (set, str) = "", keep: (set, str) = "*"):
        _drop = {c for c in drop} if isinstance(drop, str) else drop
        if keep == "*":
            _keep = {c for c in "tfgmas" if c not in _drop}
        else:
            _keep = {c for c in keep if c not in _drop} if isinstance(keep, str) else keep
        self._token_idx = set()
        if "t" in _keep:
            self._token_idx |= set(self._text_tokens)
        if "f" in _keep:
            self._token_idx |= set(self._formula_tokens)
        if "g" in _keep:
            self._token_idx |= set(self._figure_tokens)
        if "m" in _keep:
            self._token_idx |= set(self._ques_mark_tokens)
        if "a" in _keep:
            self._token_idx |= set(self._tag_tokens)
        if "s" in _keep:
            self._token_idx |= set(self._sep_tokens)
        yield
        self._token_idx = None

    def describe(self):
        return {
            "t": len(self._text_tokens),
            "f": len(self._formula_tokens),
            "g": len(self._figure_tokens),
            "m": len(self._ques_mark_tokens),
        }


def tokenize(segment_list: SegmentList, text_params=None, formula_params=None, figure_params=None):
    return TokenList(segment_list, text_params, formula_params, figure_params)


def link_formulas(*token_list: TokenList, link_vars=True):
    ast_formulas = []
    for tl in token_list:
        if tl.formula_tokenize_method == "ast":
            ast_formulas.extend([token for token in tl.inner_formula_tokens if isinstance(token, Formula)])
    _link_formulas(*ast_formulas, link_vars=link_vars)



# coding: utf-8
# 2021/5/20 @ tongshiwei

import os
from pathlib import PurePath


def abs_current_dir(filepath):
    """
    获取文件所在目录的绝对路径

    Example
    -------
    .. code ::

        abs_current_dir(__file__)

    """
    return os.path.abspath(os.path.dirname(filepath))


def path_append(path, *addition, to_str=False):
    """
    路径合并函数

    Examples
    --------
    .. code-block:: python

        path_append("../", "../data", "../dataset1/", "train", to_str=True)
        '../../data/../dataset1/train'

    Parameters
    ----------
    path: str or PurePath
    addition: list(str or PurePath)
    to_str: bool
        Convert the new path to str
    Returns
    -------

    """
    path = PurePath(path)
    if addition:
        for a in addition:
            path = path / a
    if to_str:
        return str(path)
    return path



# coding: utf-8
# 2021/5/18 @ tongshiwei

TEXT_SYMBOL = "[TEXT]"
FORMULA_SYMBOL = "[FORMULA]"
FIGURE_SYMBOL = "[FIGURE]"
QUES_MARK_SYMBOL = "[MARK]"
TAG_SYMBOL = "[TAG]"
SEP_SYMBOL = "[SEP]"
TEXT_BEGIN = r"[TEXT_BEGIN]"
TEXT_END = r"[TEXT_END]"
FORMULA_BEGIN = r"[FORMULA_BEGIN]"
FORMULA_END = r"[FORMULA_END]"


class Symbol(str):
    pass



# coding: utf-8
# 2021/7/12 @ tongshiwei

__all__ = ["PadSequence", "pad_sequence"]


class PadSequence(object):
    """Pad the sequence.

    Pad the sequence to the given `length` by inserting `pad_val`. If `clip` is set,
    sequence that has length larger than `length` will be clipped.

    Parameters
    ----------
    length : int
        The maximum length to pad/clip the sequence
    pad_val : number
        The pad value. Default 0
    clip : bool
    """

    def __init__(self, length, pad_val=0, clip=True):
        self._length = length
        self._pad_val = pad_val
        self._clip = clip

    def __call__(self, sample: list):
        """

        Parameters
        ----------
        sample : list of number

        Returns
        -------
        ret : list of number
        """
        sample_length = len(sample)
        if sample_length >= self._length:
            if self._clip and sample_length > self._length:
                return sample[:self._length]
            else:
                return sample
        else:
            return sample + [
                self._pad_val for _ in range(self._length - sample_length)
            ]


def pad_sequence(sequence: list, max_length=None, pad_val=0, clip=True):
    """

    Parameters
    ----------
    sequence
    max_length
    pad_val
    clip

    Returns
    -------

    Examples
    --------
    >>> seq = [[4, 3, 3], [2], [3, 3, 2]]
    >>> pad_sequence(seq)
    [[4, 3, 3], [2, 0, 0], [3, 3, 2]]
    >>> pad_sequence(seq, pad_val=1)
    [[4, 3, 3], [2, 1, 1], [3, 3, 2]]
    >>> pad_sequence(seq, max_length=2)
    [[4, 3], [2, 0], [3, 3]]
    >>> pad_sequence(seq, max_length=2, clip=False)
    [[4, 3, 3], [2, 0], [3, 3, 2]]
    """
    padder = PadSequence(max([len(seq) for seq in sequence]) if max_length is None else max_length, pad_val, clip)
    return [padder(seq) for seq in sequence]



# coding: utf-8
# 2021/5/20 @ tongshiwei

from EduNLP.SIF import is_sif
from EduNLP.SIF import to_sif
from EduNLP.SIF import sif4sci
import pytest


def test_is_sif():
    text = '若$x,y$满足约束条件' \
           '$\\left\\{\\begin{array}{c}2 x+y-2 \\leq 0 \\\\ x-y-1 \\geq 0 \\\\ y+1 \\geq 0\\end{array}\\right.$，' \
           '则$z=x+7 y$的最大值$\\SIFUnderline$'
    assert is_sif(text) == 1

    text = '公式需要满足完整性，完整的公式如' \
           '$\\begin{matrix} a & b \\\\ c & d \\end{matrix}$' \
           '，不完整的公式如$\\begin{matrix} a & b \\\\ c & d$'
    with pytest.raises(ValueError):
        is_sif(text)

    text = '公式需要满足符合katex的支持性，可支持的公式如' \
           '$\\begin{matrix} a & b \\\\ c & d \\end{matrix}$' \
           '，不可支持的公式如$\\frac{ \\dddot y }{ x }$'
    with pytest.raises(ValueError):
        is_sif(text)


def test_to_sif():
    text = '某校一个课外学习小组为研究某作物的发芽率y和温度x（单位...'
    siftext = to_sif(text)
    print(siftext)


def test_sci4sif(figure0, figure1, figure0_base64, figure1_base64):
    repr(sif4sci(
        r"如图所示，则$\bigtriangleup ABC$的面积是$\SIFBlank$。$\FigureID{1}$",
        tokenization_params={
            "formula_params": {
                "method": "ast",
                "return_type": "ast"
            }
        }
    ))
    repr(sif4sci(
        r"如图所示，则$\FormFigureID{0}$的面积是$\SIFBlank$。$\FigureID{1}$",
        figures={
            "0": figure0,
            "1": figure1
        },
    ))
    repr(sif4sci(
        item=r"如图所示，则$\FormFigureBase64{%s}$的面积是$\SIFBlank$。$\FigureBase64{%s}$" % (
            figure0_base64, figure1_base64
        ),
        tokenization_params={
            "figure_params": {"figure_instance": True}
        }
    ))



# coding: utf-8
# 2021/8/2 @ tongshiwei
import pytest
from EduNLP import get_pretrained_i2v
from EduNLP.Vector.t2v import PRETRAINED_MODELS
from EduNLP.I2V.i2v import MODELS
from EduNLP.I2V import D2V, W2V


def test_pretrained_i2v(tmp_path):
    PRETRAINED_MODELS["test_d2v"] = ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/d2v/test_256.zip", "d2v"]
    MODELS["test_d2v"] = [D2V, "test_d2v"]

    d = tmp_path / "model"
    d.mkdir()

    get_pretrained_i2v("test_d2v", d)

    with pytest.raises(KeyError):
        get_pretrained_i2v("error")

    PRETRAINED_MODELS["test_w2v"] = ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/w2v/test_w2v_256.zip", "w2v"]
    MODELS["test_w2v"] = [W2V, "test_w2v"]

    get_pretrained_i2v("test_w2v", d)

    with pytest.raises(KeyError):
        get_pretrained_i2v("error")



# coding: utf-8
# 2021/8/1 @ tongshiwei

import os
from os.path import expanduser, join

ROOT = os.environ.get("EDUNLPPATH", join(expanduser("~"), ".EduNLP"))
MODEL_DIR = os.environ.get("EDUNLPMODELPATH", join(ROOT, "model"))



# coding: utf-8
# 2021/5/30 @ tongshiwei

from contextlib import contextmanager

ann_format = r"$\SIFTag{{{}}}$"
ann_begin_format = r"$\SIFTag{{{}_begin}}$"
ann_end_format = r"$\SIFTag{{{}_end}}$"
ann_list_no_format = r"$\SIFTag{{list_{}}}$"


@contextmanager
def add_annotation(key, tag_mode, tar: list, key_as_tag=True):
    if key_as_tag is True:
        if tag_mode == "delimiter":
            tar.append(ann_begin_format.format(key))
        elif tag_mode == "head":
            tar.append(ann_format.format(key))
    yield
    if key_as_tag is True:
        if tag_mode == "delimiter":
            tar.append(ann_end_format.format(key))
        elif tag_mode == "tail":
            tar.append(ann_format.format(key))


def dict2str4sif(obj: dict, key_as_tag=True, tag_mode="delimiter", add_list_no_tag=True, keys=None) -> str:
    r"""

    Parameters
    ----------
    obj
    key_as_tag
    tag_mode
        delimiter: add $\SIFTag{key_begin}$ in the head and add $\SIFTag{key_end}$ at the end
        head: add $\SIFTag{key}$ in the head
        tail: add $\SIFTag{key}$ at the end
    add_list_no_tag
    keys

    Examples
    -------
    >>> item = {
    ...     "stem": r"若复数$z=1+2 i+i^{3}$，则$|z|=$",
    ...     "options": ['0', '1', r'$\sqrt{2}$', '2'],
    ... }
    >>> item
    {'stem': '若复数$z=1+2 i+i^{3}$，则$|z|=$', 'options': ['0', '1', '$\\sqrt{2}$', '2']}
    >>> dict2str4sif(item) # doctest: +ELLIPSIS
    '$\\SIFTag{stem_begin}$...$\\SIFTag{stem_end}$$\\SIFTag{options_begin}$...$\\SIFTag{options_end}$'
    >>> dict2str4sif(item, add_list_no_tag=True) # doctest: +ELLIPSIS
    '...$\\SIFTag{options_begin}$$\\SIFTag{list_0}$0$\\SIFTag{list_1}$1...$\\SIFTag{options_end}$'
    >>> dict2str4sif(item, tag_mode="head") # doctest: +ELLIPSIS
    '$\\SIFTag{stem}$...$\\SIFTag{options}$...'
    >>> dict2str4sif(item, tag_mode="tail") # doctest: +ELLIPSIS
    '若复数$z=1+2 i+i^{3}$，则$|z|=$$\\SIFTag{stem}$...2$\\SIFTag{options}$'
    >>> dict2str4sif(item, add_list_no_tag=False) # doctest: +ELLIPSIS
    '...$\\SIFTag{options_begin}$0$\\SIFSep$1$\\SIFSep$...$\\SIFTag{options_end}$'
    >>> dict2str4sif(item, key_as_tag=False)
    '若复数$z=1+2 i+i^{3}$，则$|z|=$0$\\SIFSep$1$\\SIFSep$$\\sqrt{2}$$\\SIFSep$2'
    """
    ret = []
    keys = obj.keys() if keys is None else keys
    for key in keys:
        _obj = []
        value = obj[key]
        with add_annotation(key, tag_mode, _obj, key_as_tag):
            if isinstance(value, str):
                _obj.append(value)
            elif isinstance(value, (list, dict)):
                value = value.values() if isinstance(value, dict) else value
                for i, v in enumerate(value):
                    v = str(v)
                    if key_as_tag is True and add_list_no_tag is True:
                        _obj.append(ann_list_no_format.format(i))
                    else:
                        if i > 0:
                            _obj.append(r"$\SIFSep$")
                    _obj.append(v)
            else:  # pragma: no cover
                raise TypeError("Cannot handle %s" % type(value))
        ret.append("".join(_obj))
    return str("".join(ret))



__AUTHOR__ = "Xin Wang"

from enum import IntFlag
import re


def cut(formula, preserve_braces=True, with_dollar=False,
        preserve_dollar=False, number_as_tag=False, preserve_src=True):  # pragma: no cover
    class States(IntFlag):
        CHAR = 0
        MATH = 1
        TAG = 2
        ESC = 3
        COMMAND = 4
        ARG = 5
        NUMBER = 8

    rv = []
    buffer = ''
    if with_dollar:
        state = States.CHAR
    else:
        state = States.MATH

    for c in formula:
        if state & States.NUMBER:
            if c.isdigit() or c == '.':
                buffer += c
                # c is consumed, continue
                continue
            else:
                state ^= States.NUMBER
                if number_as_tag:
                    if len(buffer) == 1:
                        rv.append(buffer)
                    elif '.' in buffer:
                        rv.append('{decimal}')
                    else:
                        rv.append('{integer}')
                else:
                    rv.append(buffer)
                buffer = ''

        if state == States.COMMAND:
            if buffer == '\\begin' or buffer == '\\end':
                state = States.ARG
                rv.append(buffer)
                buffer = c
                # c is consumed, continue
                continue
            elif c.isalpha():
                buffer += c
                # c is consumed, continue
                continue
            else:
                state = States.MATH
                if len(buffer) == 1:
                    buffer += c
                    rv.append(buffer)
                    buffer = ''
                    # c is consumed, continue
                    continue
                else:
                    rv.append(buffer)
                    buffer = ''

        if state == States.ESC:
            state = States.CHAR
            rv.append('\\' + c)
        elif state == States.CHAR:
            if c == '\\':
                state = States.ESC
            elif c.isdigit():
                state |= States.NUMBER
                buffer += c
            elif c == '{':
                state = States.TAG
                buffer += c
            elif c == '$':
                if preserve_dollar:
                    rv.append(c)
                state = States.MATH
            else:
                if c != ' ':
                    rv.append(c)
        elif state == States.TAG:
            if c == '}':
                state = States.CHAR
                buffer += c
                if not preserve_src:
                    if buffer.startswith('{img'):
                        buffer = '{img}'
                rv.append(buffer)
                buffer = ''
            else:
                buffer += c
        elif state == States.MATH:
            if c == '$':
                if preserve_dollar:
                    rv.append(c)
                state = States.CHAR
            elif c == '\\':
                state = States.COMMAND
                buffer += c
            elif c.isdigit():
                state |= States.NUMBER
                buffer += c
            else:
                if preserve_braces or (c != '{' and c != '}'):
                    if c != ' ':
                        rv.append(c)
        else:  # state == State.ARG
            if c == '}':
                state = States.MATH
                buffer += c
                rv.append(buffer)
                buffer = ''
            else:
                buffer += c
    if len(buffer) > 0:
        if state == States.NUMBER:
            if number_as_tag:
                if len(buffer) == 1:
                    rv.append(buffer)
                elif '.' in buffer:
                    rv.append('{decimal}')
                else:
                    rv.append('{integer}')
            else:
                rv.extend(buffer)
        else:
            rv.append(buffer)

    return rv


def reduce(fea):  # pragma: no cover
    rules = [
        ('a r c s i n', 'arcsin'),
        ('a r c c o s', 'arccos'),
        ('a r c t a n', 'arctan'),
        ('s i n h', 'sinh'),
        ('c o s h', 'cosh'),
        ('t a n h', 'tanh'),
        ('s i n', 'sin'),
        ('c o s', 'cos'),
        ('t a n', 'tan'),
        ('c o t', 'cot'),
        ('s e c', 'sec'),
        ('c s c', 'csc'),
        ('l g', 'lg'),
        ('l o g', 'log'),
        ('l n', 'ln'),
        ('m a x', 'max'),
        ('m i n', 'min'),
        ('{ i m g }', '{img}'),
        ('i m g', '{img}'),
        ('< u >', '{blank}'),
        ('  ', ' ')
    ]
    fea = ' '.join(fea)
    for a, b in rules:
        fea = fea.replace(a, b)
    return fea.strip().split()


def connect_char(words):  # pragma: no cover
    result = []
    buffer = ""
    for w in words:
        w = w.strip()
        if len(w) > 1:
            if len(buffer) > 0:
                result.append(buffer)
                buffer = ""
            result.append(w)

        elif len(w) == 1:
            if not w.isalpha():
                if len(buffer) > 0:
                    result.append(buffer)
                buffer = ""
                result.append(w)
            else:
                buffer += w
    if len(buffer) > 0:
        result.append(buffer)
        buffer = ""
    return result


def latex_parse(formula, preserve_braces=True, with_dollar=True,
                preserve_dollar=False, number_as_tag=False, preserve_src=True):  # pragma: no cover
    # cut
    formula_cut = cut(formula, preserve_braces, with_dollar,
                      preserve_dollar, number_as_tag, preserve_src)
    formula_reduce = reduce(formula_cut)
    formula_con = connect_char(formula_reduce)
    return formula_con


def linear_tokenize(formula, preserve_braces=True, number_as_tag=False, *args, **kwargs):
    """

    Parameters
    ----------
    formula
    preserve_braces
    number_as_tag
    args
    kwargs

    Returns
    -------

    Examples
    --------
    >>> linear_tokenize(r"{x + y}^\\frac{1}{2} + 1 = 0")
    ['{', 'x', '+', 'y', '}', '^', '\\\\frac', '{', '1', '}', '{', '2', '}', '+', '1', '=', '0']
    >>> linear_tokenize(r"ABC,AB,AC")
    ['ABC', ',', 'AB', ',', 'AC']
    """
    _formula_cut = cut(formula, preserve_braces=preserve_braces, number_as_tag=number_as_tag, *args, **kwargs)
    _formula_reduce = reduce(_formula_cut)
    _formula_con = connect_char(_formula_reduce)
    return _formula_con


# if __name__ == '__main__':
#     s = r"${x + y}^\frac{1}{2} + 1 = 0$"
#     l2 = re.split(r"(\$.+?\$)", s)
#
#     formula_cut = cut(s, with_dollar=True, preserve_braces=True)
#     formula_reduce = reduce(formula_cut)
#     formula_con = connect_char(formula_reduce)
#     print("s:", s)
#     print("formula_cut", formula_cut)
#     print("ormula_reduce", formula_reduce)
#     print("formula_con", formula_con)



# coding: utf-8
# 2021/7/12 @ tongshiwei

import torch
from ..gensim_vec import W2V
from ..embedding import Embedding
from ..meta import Vector
from EduNLP.ModelZoo import rnn, set_device
from baize.torch import save_params


class RNNModel(Vector):
    """
    Examples
    --------
    >>> model = RNNModel("ELMO", None, 2, vocab_size=4, embedding_dim=3)
    >>> seq_idx = [[1, 2, 3], [1, 2, 0], [3, 0, 0]]
    >>> output, hn = model(seq_idx, indexing=False, padding=False)
    >>> seq_idx = [[1, 2, 3], [1, 2], [3]]
    >>> output, hn = model(seq_idx, indexing=False, padding=True)
    >>> output.shape
    torch.Size([3, 3, 4])
    >>> hn.shape
    torch.Size([2, 3, 2])
    >>> tokens = model.infer_tokens(seq_idx, indexing=False)
    >>> tokens.shape
    torch.Size([3, 3, 4])
    >>> tokens = model.infer_tokens(seq_idx, agg="mean", indexing=False)
    >>> tokens.shape
    torch.Size([3, 4])
    >>> item = model.infer_vector(seq_idx, indexing=False)
    >>> item.shape
    torch.Size([3, 4])
    >>> item = model.infer_vector(seq_idx, agg="mean", indexing=False)
    >>> item.shape
    torch.Size([3, 2])
    >>> item = model.infer_vector(seq_idx, agg=None, indexing=False)
    >>> item.shape
    torch.Size([2, 3, 2])
    """

    def __init__(self, rnn_type, w2v: (W2V, tuple, list, dict, None), hidden_size,
                 freeze_pretrained=True, model_params=None, device=None,
                 **kwargs):
        self.embedding = Embedding(w2v, freeze_pretrained, **kwargs)
        for key in ["vocab_size", "embedding_dim"]:
            if key in kwargs:
                kwargs.pop(key)
        self.rnn = rnn.LM(
            rnn_type,
            self.embedding.vocab_size,
            self.embedding.embedding_dim,
            hidden_size=hidden_size,
            embedding=self.embedding.embedding,
            model_params=model_params,
            **kwargs
        )
        self.bidirectional = self.rnn.rnn.bidirectional
        self.hidden_size = self.rnn.hidden_size
        self.freeze_pretrained = freeze_pretrained
        if device is not None:
            self.set_device(device)

    def __call__(self, items, indexing=True, padding=True, **kwargs):
        seq_idx, seq_len = self.embedding(items, indexing=indexing, padding=padding, vectorization=False)

        tokens, item = self.rnn(torch.LongTensor(seq_idx), torch.LongTensor(seq_len))

        return tokens, item

    def infer_vector(self, items, agg: (int, str, None) = -1, indexing=True, padding=True, *args,
                     **kwargs) -> torch.Tensor:
        vector = self(items, indexing=indexing, padding=padding, **kwargs)[1]
        if agg is not None:
            if agg == -1:
                return torch.reshape(vector, (vector.shape[1], -1))
            return eval("torch.%s" % agg)(vector, dim=0)
        return vector

    def infer_tokens(self, items, agg=None, *args, **kwargs) -> torch.Tensor:
        tokens = self(items, **kwargs)[0]
        if agg is not None:
            return eval("torch.%s" % agg)(tokens, dim=1)
        return tokens

    @property
    def vector_size(self) -> int:
        return self.hidden_size * (1 if self.bidirectional is False else 2)

    def set_device(self, device):
        self.rnn = set_device(self.rnn, device)

    def save(self, filepath, save_embedding=False):
        save_params(filepath, self.rnn, select=None if save_embedding is True else '^(?!.*embedding)')
        return filepath

    def freeze(self, *args, **kwargs):
        return self.eval()

    @property
    def is_frozen(self):
        return not self.rnn.training

    def eval(self):
        self.rnn.eval()
        return self

    def train(self, mode=True):
        self.rnn.train(mode)
        return self



# coding: utf-8
# 2021/5/30 @ tongshiwei

import torch
import numpy as np
import pytest
from EduNLP.Pretrain import train_vector, GensimWordTokenizer
from EduNLP.Vector import W2V, D2V, RNNModel, T2V, Embedding
from EduNLP.I2V import D2V as I_D2V, W2V as I_W2V
from EduNLP.Tokenizer import get_tokenizer


@pytest.fixture(scope="module")
def stem_data(data):
    _data = []
    for e in data[:10]:
        d = e["stem"]
        _data.append(d)
    assert _data
    return _data


@pytest.fixture(scope="module")
def stem_tokens(stem_data):
    _data = []
    tokenizer = GensimWordTokenizer()
    for e in stem_data:
        d = tokenizer(e)
        if d is not None:
            _data.append(d.tokens)
    assert _data
    return _data


@pytest.fixture(scope="module")
def stem_text_tokens(stem_data):
    _data = []
    tokenizer = get_tokenizer("pure_text")
    tokens = tokenizer(stem_data)
    _data = [d for d in tokens]
    assert _data
    return _data


@pytest.fixture(scope="module")
def stem_data_general(data):
    test_items = [
        {'stem': '有公式$\\FormFigureID{wrong1?}$和公式$\\FormFigureBase64{wrong2?}$，\
            如图$\\FigureID{088f15ea-8b7c-11eb-897e-b46bfc50aa29}$,若$x,y$满足约束条件$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$'},
        {'stem': '如图$\\FigureID{088f15ea-8b7c-11eb-897e-b46bfc50aa29}$, \
            若$x,y$满足约束条件$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$'}
    ]
    data = test_items + data
    _data = []
    tokenizer = GensimWordTokenizer(symbol="gmas", general=True)
    for e in data[:10]:
        d = tokenizer(e["stem"])
        if d is not None:
            _data.append(d.tokens)
    assert _data
    return _data


@pytest.mark.parametrize("method", ["sg", "cbow", "fasttext"])
@pytest.mark.parametrize("binary", [True, False, None])
def test_w2v(stem_tokens, tmpdir, method, binary):
    filepath_prefix = str(tmpdir.mkdir(method).join("stem_tf_"))
    filepath = train_vector(
        stem_tokens,
        filepath_prefix,
        10,
        method=method,
        binary=binary,
        train_params=dict(min_count=0)
    )
    w2v = W2V(filepath, method=method, binary=binary)
    assert w2v.vector_size == 10
    w2v(*stem_tokens[0])
    assert len(w2v.infer_vector([stem_tokens[0]])[0]) == w2v.vector_size
    w2v.key_to_index(stem_tokens[0][0])
    assert len(w2v) > 0
    assert len(w2v["[FIGURE]"]) == 10
    assert len(list(w2v("[FIGURE]"))) == 1
    assert np.array_equal(w2v["[UNK]"], np.zeros((10,)))
    assert np.array_equal(w2v["[PAD]"], np.zeros((10,)))
    assert w2v.vectors.shape == (len(w2v.wv.vectors) + len(w2v.constants), w2v.vector_size)
    assert w2v.key_to_index("[UNK]") == 0
    assert w2v.key_to_index("OOV") == 0

    t2v = T2V("w2v", filepath=filepath, method=method, binary=binary)
    assert len(t2v(stem_tokens[:1])[0]) == t2v.vector_size
    assert len(t2v.infer_vector(stem_tokens[:1])[0]) == t2v.vector_size
    assert len(t2v.infer_tokens(stem_tokens[:1])[0][0]) == t2v.vector_size

    for _w2v in [[filepath, method, binary], dict(filepath=filepath, method=method, binary=binary)]:
        embedding = Embedding(_w2v, device="cpu")
        items, item_len = embedding(stem_tokens[:5])
        assert items.shape == (5, max(item_len), embedding.embedding_dim)


def test_w2v_i2v(stem_text_tokens, tmpdir, stem_data):
    method = "sg"
    filepath_prefix = str(tmpdir.mkdir(method).join("stem_tf_"))
    filepath = train_vector(
        stem_text_tokens,
        filepath_prefix,
        10,
        method=method,
        train_params=dict(min_count=0)
    )

    i2v = I_W2V("pure_text", "w2v", filepath)
    i_vec, t_vec = i2v(stem_data[:1])
    assert len(i_vec[0]) == i2v.vector_size
    assert len(t_vec[0][0]) == i2v.vector_size

    cfg_path = str(tmpdir / method / "i2v_config.json")
    i2v.save(config_path=cfg_path)
    i2v = I_W2V.load(cfg_path)

    i_vec = i2v.infer_item_vector(stem_data[:1])
    assert len(i_vec[0]) == i2v.vector_size

    t_vec = i2v.infer_token_vector(stem_data[:1])
    assert len(t_vec[0][0]) == i2v.vector_size


def test_embedding():
    with pytest.raises(TypeError):
        Embedding("error")


def test_rnn(stem_tokens, tmpdir):
    method = "sg"
    filepath_prefix = str(tmpdir.mkdir(method).join("stem_tf_"))
    filepath = train_vector(
        stem_tokens,
        filepath_prefix,
        10,
        method=method,
        train_params=dict(min_count=0)
    )
    w2v = W2V(filepath, method=method)

    with pytest.raises(TypeError):
        RNNModel("Error", w2v, 20)

    for rnn_type in ["ElMo", "Rnn", "lstm", "GRU"]:
        rnn = RNNModel(rnn_type, w2v, 20, device="cpu")

        tokens = rnn.infer_tokens(stem_tokens[:1])
        item = rnn.infer_vector(stem_tokens[:1])
        assert tokens.shape == (1, len(stem_tokens[0]), 20 * (2 if rnn.bidirectional else 1))
        assert item.shape == (1, rnn.vector_size)
        item_vec = rnn.infer_vector(stem_tokens[:1])
        assert torch.equal(item, item_vec)

        t2v = T2V(rnn_type, w2v, 20)
        assert len(t2v(stem_tokens[:1])[0]) == t2v.vector_size

        saved_params = rnn.save(str((tmpdir / method).join("stem_tf_rnn.params")), save_embedding=True)

        rnn = RNNModel(rnn_type, w2v, 20, device="cpu", model_params=saved_params)
        rnn.train()
        assert rnn.is_frozen is False
        rnn.freeze()
        assert rnn.is_frozen is True
        item_vec1 = rnn.infer_vector(stem_tokens[:1])
        assert torch.equal(item, item_vec1)


def test_d2v(stem_text_tokens, tmpdir, stem_data):
    method = "d2v"
    filepath_prefix = str(tmpdir.mkdir(method).join("stem_tf_"))
    filepath = train_vector(
        stem_text_tokens,
        filepath_prefix,
        10,
        method=method,
        train_params=dict(min_count=0)
    )
    d2v = D2V(filepath)
    assert len(d2v(stem_text_tokens[0])) == 10
    assert d2v.vector_size == 10

    t2v = T2V("d2v", filepath)
    assert len(t2v(stem_text_tokens[:1])[0]) == t2v.vector_size

    i2v = I_D2V("pure_text", "d2v", filepath)
    i_vec, t_vec = i2v(stem_data[:1])
    assert len(i_vec[0]) == i2v.vector_size
    assert t_vec is None

    cfg_path = str(tmpdir / method / "i2v_config.json")
    i2v.save(config_path=cfg_path)
    i2v = I_D2V.load(cfg_path)

    i_vec = i2v.infer_item_vector(stem_data[:1])
    assert len(i_vec[0]) == i2v.vector_size

    t_vec = i2v.infer_token_vector(stem_data[:1])
    assert t_vec is None


@pytest.mark.parametrize("method", ["bow", "tfidf"])
def test_d2v_bow_tfidf(stem_data_general, tmpdir, method):
    filepath_prefix = str(tmpdir.mkdir(method).join("stem_tf_"))
    filepath = train_vector(
        stem_data_general,
        filepath_prefix,
        method=method
    )
    d2v = D2V(filepath, method=method)
    d2v(stem_data_general[0])
    assert d2v.vector_size > 0


def test_exception(stem_tokens, tmpdir):
    filepath_prefix = str(tmpdir.mkdir("error").join("stem_tf_"))
    with pytest.raises(ValueError):
        train_vector(
            stem_tokens,
            filepath_prefix,
            embedding_dim=10,
            method="error",
            train_params=dict(min_count=0)
        )
    with pytest.raises(ValueError):
        D2V("error_path", method="error")



# coding: utf-8
# 2021/5/20 @ tongshiwei
import pytest
from EduNLP.Formula import Formula, FormulaGroup


def test_formula():
    formula = r"x + x"
    f = Formula(formula)
    f.variable_standardization(inplace=False)
    f.variable_standardization(inplace=True)
    assert len(f.ast_graph.nodes) == len(f.ast)
    f.to_str()

    formula = r"\frac{\pi}{2}"
    f = Formula(formula, variable_standardization=True)
    assert repr(f) == r"<Formula: \frac{\pi}{2}>"

    f = Formula(f.ast)
    assert f.resetable is False
    with pytest.raises(TypeError):
        f.reset_ast()

    fg = FormulaGroup([r"x + x", r"x + \frac{\pi}{2}"], variable_standardization=True)
    for f in fg:
        assert f in fg
    assert len(fg[0].ast) == 3
    fg.to_str()

    fg = FormulaGroup(["x", "y", "x"])
    assert len(fg.ast) == 3

    with pytest.raises(TypeError):
        FormulaGroup([{}])



import pytest

from EduNLP.Formula.ast import str2ast


def test_ast():
    ast_str_list = []
    # normal examples
    ast_str_list.append(r"{x + y}^\frac{\pi}{2} + 1 = x")
    ast_str_list.append(r"\color{#0FF} x = y")
    ast_str_list.append(r"x^2 + 1 = y")
    ast_str_list.append(r"\verb!x^2!")
    ast_str_list.append(r"\utilde{AB}")
    ast_str_list.append(r"\mathrm{Ab0}")
    ast_str_list.append(r"{1,2,3}")
    ast_str_list.append(r"\huge AB")
    ast_str_list.append(r"\underline{AB}")
    ast_str_list.append(r"\sqrt{\smash[b]{y}}")
    ast_str_list.append(r"\hbox{AA BB}")
    ast_str_list.append(r"abc\llap{abcdefghi}")
    ast_str_list.append(r"\raisebox{3em}{hi}")
    ast_str_list.append(r"\textcolor{#228B22}{F=ma}")
    ast_str_list.append(r"\displaystyle\sum_{i=1}^n")
    ast_str_list.append(r"\def\foo{x^2} \foo + \foo")
    ast_str_list.append(r"thank \hphantom{xyz} you")
    ast_str_list.append(r"\mathchoice{D}{T}{S}{SS}")
    ast_str_list.append(r"\bigotimes")
    ast_str_list.append(r"{AB}_b^c")
    ast_str_list.append(r"\left\{\begin{array}{c}2 x+y-2 \leq 0 \\ x-y-1 \geq 0 \\ y+1 \geq 0\end{array}\right.")
    ast_str_list.append(r"\cancel{5}")

    # work only when katex is in 'display' mode :
    ast_str_list.append(r"\begin{matrix} a & b \\ c & d \end{matrix}")
    ast_str_list.append(r"\begin{pmatrix} a&b\\c&d \end{pmatrix}")
    ast_str_list.append(r"\begin{matrix}k个\\ \overbrace{(-1)^{k-1}k,\cdots,(-1)^{k-1}k}\end{matrix}")

    # work only when 'trust' katex html func:
    ast_str_list.append(r"\href{https://katex.org}{katex}")
    ast_str_list.append(r"\htmlStyle{color: red;}{x}")
    ast_str_list.append(r"\url{www.baidu.com}")
    ast_str_list.append(r"\htmlId{bar}{x}")
    ast_str_list.append(r"\htmlClass{foo}{x}")
    ast_str_list.append("\\includegraphics[height=0.8em, totalheight=0.9em, \
    width=0.9em, alt=KA logo]{https://katex.org/img/khan-academy.png}")

    for ast_str in ast_str_list:
        str2ast(ast_str)



# coding: utf-8
# 2021/5/29 @ tongshiwei

import numpy as np
from pathlib import PurePath
from gensim.models import KeyedVectors, Word2Vec, FastText, Doc2Vec, TfidfModel
from gensim import corpora
import re
from .const import UNK, PAD
from .meta import Vector


class W2V(Vector):
    def __init__(self, filepath, method=None, binary=None):
        """

        Parameters
        ----------
        filepath:
            path to the pretrained model file
        method
        binary
        """
        fp = PurePath(filepath)
        self.binary = binary if binary is not None else (True if fp.suffix == ".bin" else False)
        if self.binary is True:
            if method == "fasttext":
                self.wv = FastText.load(filepath).wv
            else:
                self.wv = Word2Vec.load(filepath).wv
        else:
            self.wv = KeyedVectors.load(filepath, mmap="r")

        self.method = method
        self.constants = {UNK: 0, PAD: 1}

    def __len__(self):
        return len(self.constants) + len(self.wv.key_to_index)

    def key_to_index(self, word):
        if word in self.constants:
            return self.constants[word]
        else:
            if word in self.wv.key_to_index:
                return self.wv.key_to_index[word] + len(self.constants)
            else:
                return self.constants[UNK]

    @property
    def vectors(self):
        return np.concatenate([np.zeros((len(self.constants), self.vector_size)), self.wv.vectors], axis=0)

    @property
    def vector_size(self):
        return self.wv.vector_size

    def __call__(self, *words):
        for word in words:
            yield self[word]

    def __getitem__(self, item):
        return self.wv[item] if item not in self.constants else np.zeros((self.vector_size,))

    def infer_vector(self, items, agg="mean", *args, **kwargs) -> np.ndarray:
        tokens = self.infer_tokens(items, *args, **kwargs)
        return eval("np.%s" % agg)(tokens, axis=1)

    def infer_tokens(self, items, *args, **kwargs) -> list:
        return [list(self(*item)) for item in items]


class BowLoader(object):
    def __init__(self, filepath):
        self.dictionary = corpora.Dictionary.load(filepath)

    def infer_vector(self, item, return_vec=False):
        item = self.dictionary.doc2bow(item)
        if not return_vec:
            return item  # return dic as default
        vec = [0 for i in range(len(self.dictionary.keys()))]
        for i, v in item:
            vec[i] = v
        return vec

    @property
    def vector_size(self):
        return len(self.dictionary.keys())


class TfidfLoader(object):
    def __init__(self, filepath):
        self.tfidf_model = TfidfModel.load(filepath)
        # 'tfidf' model shold be used based on 'bow' model
        dictionary_path = re.sub(r"(.*)tfidf", r"\1bow", filepath)
        self.dictionary = corpora.Dictionary.load(dictionary_path)

    def infer_vector(self, item, return_vec=False):
        dic_item = self.dictionary.doc2bow(item)
        tfidf_item = self.tfidf_model[dic_item]
        # return dic as default
        if not return_vec:
            return tfidf_item  # pragma: no cover
        vec = [0 for i in range(len(self.dictionary.keys()))]
        for i, v in tfidf_item:
            vec[i] = v
        return vec

    @property
    def vector_size(self):
        return len(self.dictionary.token2id)


class D2V(Vector):
    def __init__(self, filepath, method="d2v"):
        self._method = method
        self._filepath = filepath
        if self._method == "d2v":
            self.d2v = Doc2Vec.load(filepath)
        elif self._method == "bow":
            self.d2v = BowLoader(filepath)
        elif self._method == "tfidf":
            self.d2v = TfidfLoader(filepath)
        else:
            raise ValueError("Unknown method: %s" % method)

    def __call__(self, item):
        if self._method == "d2v":
            return self.d2v.infer_vector(item)
        else:
            return self.d2v.infer_vector(item, return_vec=True)

    @property
    def vector_size(self):
        if self._method == "d2v":
            return self.d2v.vector_size
        elif self._method == "bow":
            return self.d2v.vector_size
        elif self._method == "tfidf":
            return self.d2v.vector_size

    def infer_vector(self, items, *args, **kwargs) -> list:
        return [self(item) for item in items]

    def infer_tokens(self, item, *args, **kwargs) -> ...:
        raise NotImplementedError



# coding: utf-8
# 2021/5/18 @ tongshiwei

import warnings

from .linear_token import linear_tokenize
from .ast_token import ast_tokenize


def tokenize(formula, method="linear", errors="raise", **kwargs):
    """

    Parameters
    ----------
    formula
    method
    errors: how to handle the exception occurs in ast tokenize
        "coerce": use linear_tokenize
        "raise": raise exception
    kwargs

    Returns
    -------

    Examples
    --------
    >>> tokenize(r"\\frac{\\pi}{x + y} + 1 = x")
    ['\\\\frac', '{', '\\\\pi', '}', '{', 'x', '+', 'y', '}', '+', '1', '=', 'x']
    >>> tokenize(r"\\frac{\\pi}{x + y} + 1 = x", method="ast", ord2token=True)
    <Formula: \\frac{\\pi}{x + y} + 1 = x>
    >>> tokenize(r"\\frac{\\pi}{x + y} + 1 = x", method="ast", ord2token=True, return_type="list")
    ['mathord', '{ }', 'mathord', '+', 'mathord', '{ }', '\\\\frac', '+', 'textord', '=', 'mathord']
    """
    if method == "linear":
        return linear_tokenize(formula, **kwargs)
    elif method == "ast":
        try:
            return ast_tokenize(formula, **kwargs)
        except TypeError as e:  # pragma: no cover
            if errors == "coerce":
                warnings.warn("A type error is detected, linear tokenize is applied")
                return linear_tokenize(formula)
            else:
                raise e
    else:
        raise TypeError("Unknown method type: %s" % method)



# coding: utf-8
# 2021/8/3 @ tongshiwei

from copy import deepcopy
import numpy as np


class Masker(object):
    """
    Examples
    -------
    >>> masker = Masker(per=0.5, seed=10)
    >>> items = [[1, 1, 3, 4, 6], [2], [5, 9, 1, 4]]
    >>> masked_seq, mask_label = masker(items)
    >>> masked_seq
    [[1, 1, 0, 0, 6], [2], [0, 9, 0, 4]]
    >>> mask_label
    [[0, 0, 1, 1, 0], [0], [1, 0, 1, 0]]
    >>> items = [[1, 2, 3], [1, 1, 0], [2, 0, 0]]
    >>> masked_seq, mask_label = masker(items, [3, 2, 1])
    >>> masked_seq
    [[1, 0, 3], [0, 1, 0], [2, 0, 0]]
    >>> mask_label
    [[0, 1, 0], [1, 0, 0], [0, 0, 0]]
    >>> masker = Masker(mask="[MASK]", per=0.5, seed=10)
    >>> items = [["a", "b", "c"], ["d", "[PAD]", "[PAD]"], ["hello", "world", "[PAD]"]]
    >>> masked_seq, mask_label = masker(items, length=[3, 1, 2])
    >>> masked_seq
    [['a', '[MASK]', 'c'], ['d', '[PAD]', '[PAD]'], ['hello', '[MASK]', '[PAD]']]
    >>> mask_label
    [[0, 1, 0], [0, 0, 0], [0, 1, 0]]
    """

    def __init__(self, mask: (int, str, ...) = 0, per=0.2, seed=None):
        """

        Parameters
        ----------
        mask: int, str
        per
        seed
        """
        self.seed = np.random.default_rng(seed)
        self.per = per
        self.mask = mask

    def __call__(self, seqs, length=None, *args, **kwargs) -> tuple:
        seqs = deepcopy(seqs)
        masked_list = []
        if length is None:
            length = [len(seq) for seq in seqs]
        for seq, _length in zip(seqs, length):
            masked = self.seed.choice(len(seq) - 1, size=int(_length * self.per), replace=False)
            _masked_list = [0] * len(seq)
            for _masked in masked:
                seq[_masked] = self.mask
                _masked_list[_masked] = 1
            masked_list.append(_masked_list)
        return seqs, masked_list



# coding: utf-8
# 2021/7/12 @ tongshiwei

from typing import List
import torch
from .gensim_vec import W2V
from .const import PAD
from EduNLP.ModelZoo import pad_sequence, set_device


class Embedding(object):
    def __init__(self, w2v: (W2V, tuple, list, dict, None), freeze=True, device=None, **kwargs):
        if w2v is None:
            self.w2v = None
        elif isinstance(w2v, (tuple, list)):
            self.w2v = W2V(*w2v)
        elif isinstance(w2v, dict):
            self.w2v = W2V(**w2v)
        elif isinstance(w2v, W2V):
            self.w2v = w2v
        else:
            raise TypeError("w2v argument must be one of W2V, tuple, list, dict or None, now is %s" % type(w2v))

        if self.w2v is not None:
            self.vocab_size = len(self.w2v)
            self.embedding_dim = self.w2v.vector_size
        else:
            self.vocab_size = kwargs["vocab_size"]
            self.embedding_dim = kwargs["embedding_dim"]

        self.embedding = torch.nn.Embedding(self.vocab_size, self.embedding_dim)

        self.pad_val = 0
        if self.w2v is not None:
            self.embedding.from_pretrained(torch.Tensor(self.w2v.vectors), freeze)
            self.pad_val = self.w2v.constants[PAD]
        self.key_to_index = self.w2v.key_to_index if w2v is not None else lambda x: x

        if device is not None:
            self.set_device(device)

    def __call__(self, items: List[List[str]], indexing=True, padding=True, vectorization=True, *args,
                 **kwargs) -> tuple:

        items, item_len = self.indexing(items, padding=padding, indexing=indexing)
        items = self.infer_token_vector(items, indexing=False)[0] if vectorization else items
        return items, item_len

    def infer_token_vector(self, items: List[List[str]], indexing=True) -> tuple:
        items, item_len = self.indexing(items, padding=True, indexing=indexing)
        item_embedding = self.embedding(torch.LongTensor(items))
        return item_embedding, item_len

    def indexing(self, items: List[List[str]], padding=False, indexing=True) -> tuple:
        """

        Parameters
        ----------
        items: list of list of str(word/token)
        padding: bool
            whether padding the returned list with default pad_val to make all item in items have the same length
        indexing: bool

        Returns
        -------
        token_idx: list of list of int
            the list of the tokens of each item
        token_len: list of int
            the list of the length of tokens of each item
        """
        items_idx = [[self.key_to_index(word) for word in item] for item in items] if indexing else items
        item_len = [len(_idx) for _idx in items_idx]
        padded_items_idx = pad_sequence(items_idx, pad_val=self.pad_val) if padding is True else items_idx
        return padded_items_idx, item_len

    def set_device(self, device):
        self.embedding = set_device(self.embedding, device)
        return self



# coding: utf-8
# 2021/5/20 @ tongshiwei

import pytest

from EduNLP.SIF.segment import seg
from EduNLP.utils import image2base64


def test_segment(figure0, figure1, figure0_base64, figure1_base64):
    seg(
        r"如图所示，则$\FormFigureID{0}$的面积是$\SIFBlank$。$\FigureID{1}$",
        figures={
            "0": figure0,
            "1": figure1
        }
    )
    s = seg(
        r"如图所示，则$\FormFigureBase64{%s}$的面积是$\SIFBlank$。$\FigureBase64{%s}$" % (figure0_base64, figure1_base64),
        figures=True
    )
    with pytest.raises(TypeError):
        s.append("123")
    seg_test_text = seg(
        r"如图所示，有三组$\textf{机器人,bu}$在踢$\textf{足球,b}$",
        figures=True
    )
    assert seg_test_text.text_segments == ['如图所示，有三组机器人在踢足球']



# coding: utf-8
# 2021/5/18 @ tongshiwei

from .segment import (SegmentList, TextSegment, FigureFormulaSegment, LatexFormulaSegment, FigureSegment,
                      QuesMarkSegment, Figure, TagSegment, SepSegment, seg)



# coding: utf-8
# 2021/8/1 @ tongshiwei

from typing import Iterable
from ..SIF.segment import seg
from ..SIF.tokenization import tokenize

__all__ = ["TOKENIZER", "Tokenizer", "PureTextTokenizer", "TextTokenizer", "get_tokenizer"]


class Tokenizer(object):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class PureTextTokenizer(Tokenizer):
    r"""

    Examples
    --------
    >>> tokenizer = PureTextTokenizer()
    >>> items = ["有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"]
    >>> tokens = tokenizer(items)
    >>> next(tokens)[:10]
    ['公式', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[SEP]', 'z']
    >>> items = ["已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$"]
    >>> tokens = tokenizer(items)
    >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
    ['已知', '集合', 'A', '=', '\\left', '\\{', 'x', '\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', '4', '<',
    '0', '\\right', '\\}', ',', '\\quad', 'B', '=', '\\{', '-', '4', ',', '1', ',', '3', ',', '5', '\\}', ',',
    '\\quad', 'A', '\\cap', 'B', '=']
    >>> items = [{
    ... "stem": "已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$",
    ... "options": ["1", "2"]
    ... }]
    >>> tokens = tokenizer(items, key=lambda x: x["stem"])
    >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
    ['已知', '集合', 'A', '=', '\\left', '\\{', 'x', '\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', '4', '<',
    '0', '\\right', '\\}', ',', '\\quad', 'B', '=', '\\{', '-', '4', ',', '1', ',', '3', ',', '5', '\\}', ',',
    '\\quad', 'A', '\\cap', 'B', '=']
    """

    def __init__(self, *args, **kwargs):
        self.tokenization_params = {
            "formula_params": {
                "method": "linear",
                "skip_figure_formula": True
            }
        }

    def __call__(self, items: Iterable, key=lambda x: x, *args, **kwargs):
        for item in items:
            yield tokenize(seg(key(item), symbol="gmas"), **self.tokenization_params).tokens


class TextTokenizer(Tokenizer):
    r"""

    Examples
    ----------
    >>> tokenizer = TextTokenizer()
    >>> items = ["有公式$\\FormFigureID{wrong1?}$，如图$\\FigureID{088f15ea-xxx}$,\
    ... 若$x,y$满足约束条件公式$\\FormFigureBase64{wrong2?}$,$\\SIFSep$，则$z=x+7 y$的最大值为$\\SIFBlank$"]
    >>> tokens = tokenizer(items)
    >>> next(tokens)[:10]
    ['公式', '[FORMULA]', '如图', '[FIGURE]', 'x', ',', 'y', '约束条件', '公式', '[FORMULA]']
    """

    def __init__(self, *args, **kwargs):
        self.tokenization_params = {
            "formula_params": {
                "method": "linear",
                "symbolize_figure_formula": True
            }
        }

    def __call__(self, items: Iterable, key=lambda x: x, *args, **kwargs):
        for item in items:
            yield tokenize(seg(key(item), symbol="gmas"), **self.tokenization_params).tokens


TOKENIZER = {
    "pure_text": PureTextTokenizer,
    "text": TextTokenizer
}


def get_tokenizer(name, *args, **kwargs):
    r"""

    Parameters
    ----------
    name: str
    args
    kwargs

    Returns
    -------
    tokenizer: Tokenizer

    Examples
    --------
    >>> items = ["已知集合$A=\\left\\{x \\mid x^{2}-3 x-4<0\\right\\}, \\quad B=\\{-4,1,3,5\\}, \\quad$ 则 $A \\cap B=$"]
    >>> tokenizer = get_tokenizer("text")
    >>> tokens = tokenizer(items)
    >>> next(tokens)  # doctest: +NORMALIZE_WHITESPACE
    ['已知', '集合', 'A', '=', '\\left', '\\{', 'x', '\\mid', 'x', '^', '{', '2', '}', '-', '3', 'x', '-', '4', '<',
    '0', '\\right', '\\}', ',', '\\quad', 'B', '=', '\\{', '-', '4', ',', '1', ',', '3', ',', '5', '\\}', ',',
    '\\quad', 'A', '\\cap', 'B', '=']
    """
    if name not in TOKENIZER:
        raise KeyError(
            "Unknown tokenizer %s, use one of the provided tokenizers: %s" % (name, ", ".join(TOKENIZER.keys()))
        )
    return TOKENIZER[name](*args, **kwargs)



# coding: utf-8
# 2021/8/2 @ tongshiwei
import logging
import torch
from torch.nn import DataParallel


def set_device(_net, ctx, *args, **kwargs):  # pragma: no cover
    """code from longling v1.3.26"""
    if ctx == "cpu":
        if not isinstance(_net, DataParallel):
            _net = DataParallel(_net)
        return _net.cpu()
    elif any(map(lambda x: x in ctx, ["cuda", "gpu"])):
        if not torch.cuda.is_available():
            try:
                torch.ones((1,), device=torch.device("cuda: 0"))
            except AssertionError as e:
                raise TypeError("no cuda detected, noly cpu is supported, the detailed error msg:%s" % str(e))
        if torch.cuda.device_count() >= 1:
            if ":" in ctx:
                ctx_name, device_ids = ctx.split(":")
                assert ctx_name in ["cuda", "gpu"], "the equipment should be 'cpu', 'cuda' or 'gpu', now is %s" % ctx
                device_ids = [int(i) for i in device_ids.strip().split(",")]
                try:
                    if not isinstance(_net, DataParallel):
                        return DataParallel(_net, device_ids).cuda
                    return _net.cuda(device_ids)
                except AssertionError as e:
                    logging.error(device_ids)
                    raise e
            elif ctx in ["cuda", "gpu"]:
                if not isinstance(_net, DataParallel):
                    _net = DataParallel(_net)
                return _net.cuda()
            else:
                raise TypeError("the equipment should be 'cpu', 'cuda' or 'gpu', now is %s" % ctx)
        else:
            logging.error(torch.cuda.device_count())
            raise TypeError("0 gpu can be used, use cpu")
    else:
        if not isinstance(_net, DataParallel):
            return DataParallel(_net, device_ids=ctx).cuda()
        return _net.cuda(ctx)
